class DataAggregatorAbstractError(Exception):
    """
    Ошибки которые наследуются от этого класса используются
    - для отделения ошибок НАШЕГО приложения от всех других.
    АБСТРАКТНЫЙ КЛАСС = НАПРЯМУЮ НЕ ИСПОЛЬЗУЕТСЯ для raise
    """
    def __init__(self, message: str, error: Exception, status_code: int) -> None:
        assert isinstance(message, str), f'message must be str. "{type(message).__name__}" was given'
        assert isinstance(error, Exception), f'error must be Exception. "{type(error).__name__}" was given'
        super(DataAggregatorAbstractError, self).__init__(f'{message} :: {str(error)} :: {status_code})')
        self.status_code = status_code
        self.error = error


class DataAggregatorRequestError(DataAggregatorAbstractError):
    pass


class DataAggregatorResponseError(DataAggregatorAbstractError):
    pass



from unipipeline.modules.uni import Uni

from data_aggregator_sdk.constants.names import WORKER_NAME
from data_aggregator_sdk.data_aggregator_sdk_config import DataAggregatorSdkConfig
from data_aggregator_sdk.integration_message import IntegrationV0MessageGateway
from data_aggregator_sdk.lib import CONFIG_FILE
from data_aggregator_sdk.runtime_conf import set_broker_uri


class DataAggregatorSdk:

    def __init__(self, config: DataAggregatorSdkConfig) -> None:
        self._config = config
        self._uni = Uni(CONFIG_FILE)
        set_broker_uri(self._config.broker_url)

    def init_uni_worker(self) -> None:
        self._uni.init_producer_worker(WORKER_NAME)
        self._uni.initialize()

    def send_message(self, message: IntegrationV0MessageGateway) -> None:
        self._uni.send_to(WORKER_NAME, message)



from unipipeline.brokers.uni_amqp_py_broker import UniAmqpPyBroker

from data_aggregator_sdk.runtime_conf import get_broker_uri


class DataAggregatorInputBroker(UniAmqpPyBroker):

    @classmethod
    def get_connection_uri(cls) -> str:
        return get_broker_uri()



from datetime import datetime
from typing import Dict, List, Optional, Union
from uuid import UUID


TChannelInfoBox = List[Dict[str, Union[List[Dict[str, Optional[datetime]]], int]]]

TDevicesInfoBoxItem = Dict[str, Union[int, TChannelInfoBox]]

TDevicesInfoBox = Dict[UUID, TDevicesInfoBoxItem]



from enum import Enum, unique


class NetworkTypeEnum(Enum):
    """
    Enumeration of network types
    """
    input = "input"
    output = "output"

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class DeviceModificationTypeEnum(Enum):
    """
    Enumeration of modification types for devices
    """
    smart_meter = "smart_meter"
    modem = "modem"

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class NetworkSysTypeEnum(Enum):  # TODO: move to reception sdk
    """
    Enumeration of system types for application networks
    """
    OUTPUT_OLD_LK = "OUTPUT_OLD_LK"
    OUTPUT_NEW_LK = "OUTPUT_NEW_LK"
    OUTPUT_DATA_LOGGER_DEVICE_DATA = "OUTPUT_DATA_LOGGER_DEVICE_DATA"
    OUTPUT_DATA_AGGREGATOR_DEVICE_DATA = "OUTPUT_DATA_AGGREGATOR_DEVICE_DATA"
    INPUT_NBIOT = "INPUT_NBIOT"
    INPUT_LORA = "INPUT_LORA"
    INPUT_KAFKA_BS0 = "INPUT_KAFKA_BS0"
    INPUT_HTTP_BS = "INPUT_HTTP_BS"
    INPUT_NEW_LK = "INPUT_NEW_LK"

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class ProtocolEnum(Enum):
    """
    Enumeration of protocols for packets
    """
    WATER5_V_NERO_V0 = 'WATER5_V_NERO_V0'
    NCP_SMP_V0 = 'NCP_SMP_V0'
    SMP_V0 = 'SMP_V0'
    SMP_M_GAS_METER_V0 = 'SMP_M_GAS_METER_V0'
    SMP_M_ENERGY_METER_V0 = 'SMP_M_ENERGY_METER_V0'
    SMP_M_JUPITER_08B_V0 = 'SMP_M_JUPITER_08B_V0'
    SMP_M_JUPITER_12B_V0 = 'SMP_M_JUPITER_12B_V0'
    SMP_M_JUPITER_16B_V0 = 'SMP_M_JUPITER_16B_V0'
    SMP_M_WATER_METER_04B_V0 = 'SMP_M_WATER_METER_04B_V0'
    SMP_M_WATER_METER_08B_V0 = 'SMP_M_WATER_METER_08B_V0'
    SMP_M_WATER_METER_12B_V0 = 'SMP_M_WATER_METER_12B_V0'
    SMP_M_WATER_METER_16B_V0 = 'SMP_M_WATER_METER_16B_V0'
    SMP_M_HEAT_PROXY_METER_16B_V0 = 'SMP_M_HEAT_PROXY_METER_16B_V0'
    WATER5_V_JUPITER_FREESCALE_V0 = 'WATER5_V_JUPITER_FREESCALE_V0'
    WATER5_V_JUPITER_STM_V0 = 'WATER5_V_JUPITER_STM_V0'
    WATER5_V_FLUO_STM_V0 = 'WATER5_V_FLUO_STM_V0'
    WATER5_V_FLUO_FREESCALE_V0 = 'WATER5_V_FLUO_FREESCALE_V0'
    WATER5_V_FLUO_A_V0 = 'WATER5_V_FLUO_A_V0'
    WATER5_V_FLUO_S_V0 = 'WATER5_V_FLUO_S_V0'
    WATER5_V_GAS_V0 = 'WATER5_V_GAS_V0'
    WATER5_V_JUPITER_LORA_V0 = 'WATER5_V_JUPITER_LORA_V0'
    WATER5_V_FLUO_LORA_V0 = 'WATER5_V_FLUO_LORA_V0'
    SMP_M_INTERNAL_INFO_DATA = 'SMP_M_INTERNAL_INFO_DATA'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class IntegrationV0MessageErrorType(Enum):
    """
    Enumeration of error types
    """
    none = 'none'
    device_unidentified = 'device_unidentified'
    data_undecryptable = 'data_undecryptable'
    data_unparsable = 'data_unparsable'
    mac_duplicated = 'mac_duplicated'
    packet_from_the_future = 'packet_from_the_future'
    packet_from_the_past = 'packet_from_the_past'
    error_unidentified = 'error_unidentified'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class EncryptionType(Enum):  # DO NOT CHANGE, DEPENDENCY BY RECEPTION INTERFACE!!!!!!!
    """
    Enumeration of encryption types
    """
    NO_ENCRYPTION = 'NO_ENCRYPTION'
    XTEA_V_NERO_V0 = 'XTEA_V_NERO_V0'
    AES_ECB_V_NERO_V0 = 'AES_ECB_V_NERO_V0'
    KUZNECHIK_V_NERO_V0 = 'KUZNECHIK_V_NERO_V0'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class ResourceKind(Enum):
    """
    Enumeration of resource kinds for devices
    """
    COMMON_CONSUMED = 'COMMON_CONSUMED'
    COMMON_GENERATED = 'COMMON_GENERATED'
    COMMON_ACTIVE_GENERATED = 'COMMON_ACTIVE_GENERATED'
    COMMON_ACTIVE_CONSUMED = 'COMMON_ACTIVE_CONSUMED'
    COMMON_REACTIVE_GENERATED = 'COMMON_REACTIVE_GENERATED'
    COMMON_REACTIVE_CONSUMED = 'COMMON_REACTIVE_CONSUMED'
    PHASE_A_ACTIVE_CONSUMED = 'PHASE_A_ACTIVE_CONSUMED'
    PHASE_A_ACTIVE_GENERATED = 'PHASE_A_ACTIVE_GENERATED'
    PHASE_A_REACTIVE_CONSUMED = 'PHASE_A_REACTIVE_CONSUMED'
    PHASE_A_REACTIVE_GENERATED = 'PHASE_A_REACTIVE_GENERATED'
    PHASE_B_ACTIVE_CONSUMED = 'PHASE_B_ACTIVE_CONSUMED'
    PHASE_B_ACTIVE_GENERATED = 'PHASE_B_ACTIVE_GENERATED'
    PHASE_B_REACTIVE_CONSUMED = 'PHASE_B_REACTIVE_CONSUMED'
    PHASE_B_REACTIVE_GENERATED = 'PHASE_B_REACTIVE_GENERATED'
    PHASE_C_ACTIVE_CONSUMED = 'PHASE_C_ACTIVE_CONSUMED'
    PHASE_C_ACTIVE_GENERATED = 'PHASE_C_ACTIVE_GENERATED'
    PHASE_C_REACTIVE_CONSUMED = 'PHASE_C_REACTIVE_CONSUMED'
    PHASE_C_REACTIVE_GENERATED = 'PHASE_C_REACTIVE_GENERATED'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class DeviceValueMarker(Enum):
    """
    Enumeration of markers for device values
    """
    NO_DOUBT = 'NO_DOUBT'
    OVERFLOW = 'OVERFLOW'
    OVERFLOW_SUSPICIOUS = 'OVERFLOW_SUSPICIOUS'
    NOT_CHECKED = 'NOT_CHECKED'
    REJECTED_OVERFLOW = 'REJECTED_OVERFLOW'
    REJECTED_VALUE = 'REJECTED_VALUE'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


@unique
class IntegrationV0MessageEvent(Enum):
    """
    Enumeration of events for devices
    """
    BATTERY_IS_LOW = 'BATTERY_IS_LOW'
    MAGNET_WAS_DETECTED = 'MAGNET_WAS_DETECTED'
    CASE_WAS_OPENED = 'CASE_WAS_OPENED'
    TEMPERATURE_LIMIT = 'TEMPERATURE_LIMIT'
    OTHER = 'OTHER'
    SYS_NO_DATA = 'SYS_NO_DATA'

    FLOW_REVERSE = 'FLOW_REVERSE'
    FLOW_SPEED_OVER_LIMIT = 'FLOW_SPEED_OVER_LIMIT'
    CONTINUES_CONSUMPTION = 'CONTINUES_CONSUMPTION'
    NO_WATER = 'NO_WATER'  # don't use
    NO_RESOURCE = 'NO_RESOURCE'
    BATTERY_WARNING = 'BATTERY_WARNING'
    BATTERY_OR_TEMPERATURE_LIMITS = 'BATTERY_OR_TEMPERATURE_LIMITS'

    RESET = 'RESET'  # REASON IS UNAVAILABLE
    RESET_POWER_ON = 'RESET_POWER_ON'
    RESET_PIN = 'RESET_PIN'
    RESET_LOW_VOLTAGE = 'RESET_LOW_VOLTAGE'
    RESET_SOFTWARE = 'RESET_SOFTWARE'
    RESET_WATCHDOG = 'RESET_WATCHDOG'
    RESET_HARD_FAULT = 'RESET_HARD_FAULT'

    ERROR = 'ERROR'  # REASON IS UNAVAILABLE
    ERROR_SENSOR = 'ERROR_SENSOR'
    ERROR_SYSTEM = 'ERROR_SYSTEM'
    ERROR_SENSOR_MEASUREMENT = 'ERROR_SENSOR_MEASUREMENT'
    ERROR_SENSOR_TEMPERATURE = 'ERROR_SENSOR_TEMPERATURE'
    ERROR_MEASUREMENT = 'ERROR_MEASUREMENT'
    ERROR_METER_SYNC = 'ERROR_METER_SYNC'
    ERROR_LOW_VOLTAGE = 'ERROR_LOW_VOLTAGE'
    ERROR_INTERNAL_CLOCK = 'ERROR_INTERNAL_CLOCK'
    ERROR_FLASH = 'ERROR_FLASH'
    ERROR_EEPROM = 'ERROR_EEPROM'
    ERROR_RADIO = 'ERROR_RADIO'
    ERROR_DISPLAY = 'ERROR_DISPLAY'
    ERROR_PLC = 'ERROR_PLC'
    ERROR_RESET = 'ERROR_RESET'
    IMPACT_POWER_LOST = 'IMPACT_POWER_LOST'
    IMPACT_MAGNET = 'IMPACT_MAGNET'
    IMPACT_CLEAT_TAMPER = 'IMPACT_CLEAT_TAMPER'
    IMPACT_RADIO = 'IMPACT_RADIO'

    # Journal event items
    NONE = 'NONE'
    SUCCESSFUL_AUTO_DIAGNOSTIC = 'SUCCESSFUL_AUTO_DIAGNOSTIC'
    SETUP_UPDATE = 'SETUP_UPDATE'
    SWITCH_WINTER_DAYLIGHT = 'SWITCH_WINTER_DAYLIGHT'
    SWITCH_SUMMER_DAYLIGHT = 'SWITCH_SUMMER_DAYLIGHT'
    RECORD_DATETIME = 'RECORD_DATETIME'
    CHANGE_OFFSET_DAILY_CLOCK = 'CHANGE_OFFSET_DAILY_CLOCK'
    PERMISSION_SWITCH_DAYLIGHT_ON = 'PERMISSION_SWITCH_DAYLIGHT_ON'
    PERMISSION_SWITCH_DAYLIGHT_OFF = 'PERMISSION_SWITCH_DAYLIGHT_OFF'
    CHANGE_DATE_TIME_SWITCH_DAYLIGHT = 'CHANGE_DATE_TIME_SWITCH_DAYLIGHT'
    ERASE_EEPROM = 'ERASE_EEPROM'
    NULLIFY_TARIFF_ACCUMULATION = 'NULLIFY_TARIFF_ACCUMULATION'
    NULLIFY_INTERVAL_ACCUMULATION = 'NULLIFY_INTERVAL_ACCUMULATION'
    RESET_PASSWORD = 'RESET_PASSWORD'
    RESET_POWER_LOST_TIME_COUNTER = 'RESET_POWER_LOST_TIME_COUNTER'
    RESET_MAGNET_IMPACT_TIME_COUNTER = 'RESET_MAGNET_IMPACT_TIME_COUNTER'
    RESET_POWER_INCREASE_TIME_COUNTER = 'RESET_POWER_INCREASE_TIME_COUNTER'
    RESET_POWER_DECREASE_TIME_COUNTER = 'RESET_POWER_DECREASE_TIME_COUNTER'
    RESET_MAINTS_FREQ_DIVERGENCE_TIME_COUNTER = 'RESET_MAINTS_FREQ_DIVERGENCE_TIME_COUNTER'
    RESET_POWER_OVER_LIMIT_TIME_COUNTER = 'RESET_POWER_OVER_LIMIT_TIME_COUNTER'
    CHANGE_CAPACITY_DATA_LCD = 'CHANGE_CAPACITY_DATA_LCD'
    CHANGE_TARIFF_METHODS = 'CHANGE_TARIFF_METHODS'
    CHANGE_TARIFF_PROGRAMS = 'CHANGE_TARIFF_PROGRAMS'
    CHANGE_ACTUAL_SEASON_SCHEDULES = 'CHANGE_ACTUAL_SEASON_SCHEDULES'
    CHANGE_CONSUMPTION_LIMIT = 'CHANGE_CONSUMPTION_LIMIT'
    CHANGE_LOW_THRESHOLD_VOLTAGE = 'CHANGE_LOW_THRESHOLD_VOLTAGE'
    CHANGE_HIGH_THRESHOLD_VOLTAGE = 'CHANGE_HIGH_THRESHOLD_VOLTAGE'
    CHANGE_MAINTS_FREQ_THRESHOLD = 'CHANGE_MAINTS_FREQ_THRESHOLD'
    CHANGE_THRESHOLD_LOW_CONSUMPTION = 'CHANGE_THRESHOLD_LOW_CONSUMPTION'
    RECHARGE_ENERGY_PAYMENT = 'RECHARGE_ENERGY_PAYMENT'
    UNSUCCESSFUL_AUTO_DIAGNOSTIC_INTERNAL_CLOCK = 'UNSUCCESSFUL_AUTO_DIAGNOSTIC_INTERNAL_CLOCK'
    ABNORMAL_COUNTER_AUTOSTART = 'ABNORMAL_COUNTER_AUTOSTART'
    EXTERNAL_POWER_LOST = 'EXTERNAL_POWER_LOST'
    EXTERNAL_POWER_DETECTED = 'EXTERNAL_POWER_DETECTED'
    START_POWER_OVER_LIMIT = 'START_POWER_OVER_LIMIT'
    STOP_POWER_OVER_LIMIT = 'STOP_POWER_OVER_LIMIT'
    ENERGY_OVER_LIMIT_1 = 'ENERGY_OVER_LIMIT_1'
    ENERGY_OVER_LIMIT_2 = 'ENERGY_OVER_LIMIT_2'
    ENERGY_OVER_LIMIT_3 = 'ENERGY_OVER_LIMIT_3'
    WRONG_PASSWORD_BLOCK = 'WRONG_PASSWORD_BLOCK'
    WRONG_PASSWORD_APPEAL = 'WRONG_PASSWORD_APPEAL'
    EXHAUST_DAILY_BATTERY_LIFE_LIMIT = 'EXHAUST_DAILY_BATTERY_LIFE_LIMIT'
    START_MAGNET_IMPACT = 'START_MAGNET_IMPACT'
    STOP_MAGNET_IMPACT = 'STOP_MAGNET_IMPACT'
    VIOLATION_TERMINAL_BLOCK_SEAL = 'VIOLATION_TERMINAL_BLOCK_SEAL'
    RECOVERY_TERMINAL_BLOCK_SEAL = 'RECOVERY_TERMINAL_BLOCK_SEAL'
    VIOLATION_CASE_SEAL = 'VIOLATION_CASE_SEAL'
    RECOVERY_CASE_SEAL = 'RECOVERY_CASE_SEAL'
    TIME_OUT_SYNC_LIMIT = 'TIME_OUT_SYNC_LIMIT'
    CRITICAL_DIVERGENCE_TIME = 'CRITICAL_DIVERGENCE_TIME'
    OVERHEAT_COUNTER_START = 'OVERHEAT_COUNTER_START'
    OVERHEAT_COUNTER_STOP = 'OVERHEAT_COUNTER_STOP'
    UNSUCCESSFUL_AUTO_DIAGNOSTIC_MEMORY = 'UNSUCCESSFUL_AUTO_DIAGNOSTIC_MEMORY'
    LOW_BATTERY_CAPACITY = 'LOW_BATTERY_CAPACITY'
    RECOVERY_BATTERY_WORKING_VOLTAGE = 'RECOVERY_BATTERY_WORKING_VOLTAGE'
    LOW_CONSUMPTION = 'LOW_CONSUMPTION'
    RESET_FLAG_LOW_CONSUMPTION = 'RESET_FLAG_LOW_CONSUMPTION'
    CHANGE_VALIDATION_SETTINGS = 'CHANGE_VALIDATION_SETTINGS'
    UNSUCCESSFUL_AUTO_DIAGNOSTIC_MEASUREMENT_BLOCK = 'UNSUCCESSFUL_AUTO_DIAGNOSTIC_MEASUREMENT_BLOCK'
    UNSUCCESSFUL_AUTO_DIAGNOSTIC_CALCULATION_BLOCK = 'UNSUCCESSFUL_AUTO_DIAGNOSTIC_CALCULATION_BLOCK'
    UNSUCCESSFUL_AUTO_DIAGNOSTIC_POWER_BLOCK = 'UNSUCCESSFUL_AUTO_DIAGNOSTIC_POWER_BLOCK'
    UNSUCCESSFUL_AUTO_DIAGNOSTIC_SCREEN = 'UNSUCCESSFUL_AUTO_DIAGNOSTIC_SCREEN'
    UNSUCCESSFUL_AUTO_DIAGNOSTIC_RADIO = 'UNSUCCESSFUL_AUTO_DIAGNOSTIC_RADIO'
    MAINS_VOLTAGE_LOST_PHASE_A_START = 'MAINS_VOLTAGE_LOST_PHASE_A_START'
    MAINS_VOLTAGE_LOST_PHASE_A_STOP = 'MAINS_VOLTAGE_LOST_PHASE_A_STOP'
    MAINS_VOLTAGE_LOST_PHASE_B_START = 'MAINS_VOLTAGE_LOST_PHASE_B_START'
    MAINS_VOLTAGE_LOST_PHASE_B_STOP = 'MAINS_VOLTAGE_LOST_PHASE_B_STOP'
    MAINS_VOLTAGE_LOST_PHASE_C_START = 'MAINS_VOLTAGE_LOST_PHASE_C_START'
    MAINS_VOLTAGE_LOST_PHASE_C_STOP = 'MAINS_VOLTAGE_LOST_PHASE_C_STOP'
    VOLTAGE_LAYDOWN_PHASE_A_START = 'VOLTAGE_LAYDOWN_PHASE_A_START'
    VOLTAGE_LAYDOWN_PHASE_A_STOP = 'VOLTAGE_LAYDOWN_PHASE_A_STOP'
    VOLTAGE_LAYDOWN_PHASE_B_START = 'VOLTAGE_LAYDOWN_PHASE_B_START'
    VOLTAGE_LAYDOWN_PHASE_B_STOP = 'VOLTAGE_LAYDOWN_PHASE_B_STOP'
    VOLTAGE_LAYDOWN_PHASE_C_START = 'VOLTAGE_LAYDOWN_PHASE_C_START'
    VOLTAGE_LAYDOWN_PHASE_C_STOP = 'VOLTAGE_LAYDOWN_PHASE_C_STOP'
    OVERVOLTAGE_PHASE_A_START = 'OVERVOLTAGE_PHASE_A_START'
    OVERVOLTAGE_PHASE_A_STOP = 'OVERVOLTAGE_PHASE_A_STOP'
    OVERVOLTAGE_PHASE_B_START = 'OVERVOLTAGE_PHASE_B_START'
    OVERVOLTAGE_PHASE_B_STOP = 'OVERVOLTAGE_PHASE_B_STOP'
    OVERVOLTAGE_PHASE_C_START = 'OVERVOLTAGE_PHASE_C_START'
    OVERVOLTAGE_PHASE_C_STOP = 'OVERVOLTAGE_PHASE_C_STOP'
    OVERCURRENT_PHASE_A_START = 'OVERCURRENT_PHASE_A_START'
    OVERCURRENT_PHASE_A_STOP = 'OVERCURRENT_PHASE_A_STOP'
    OVERCURRENT_PHASE_B_START = 'OVERCURRENT_PHASE_B_START'
    OVERCURRENT_PHASE_B_STOP = 'OVERCURRENT_PHASE_B_STOP'
    OVERCURRENT_PHASE_C_START = 'OVERCURRENT_PHASE_C_START'
    OVERCURRENT_PHASE_C_STOP = 'OVERCURRENT_PHASE_C_STOP'
    CURRENT_SUM_THRESHOLD_LOW_START = 'CURRENT_SUM_THRESHOLD_LOW_START'
    CURRENT_SUM_THRESHOLD_LOW_STOP = 'CURRENT_SUM_THRESHOLD_LOW_STOP'
    FREQ_OUT_PHASE_A_START = 'FREQ_OUT_PHASE_A_START'
    FREQ_OUT_PHASE_A_STOP = 'FREQ_OUT_PHASE_A_STOP'
    FREQ_OUT_PHASE_B_START = 'FREQ_OUT_PHASE_B_START'
    FREQ_OUT_PHASE_B_STOP = 'FREQ_OUT_PHASE_B_STOP'
    FREQ_OUT_PHASE_C_START = 'FREQ_OUT_PHASE_C_START'
    FREQ_OUT_PHASE_C_STOP = 'FREQ_OUT_PHASE_C_STOP'
    PHASE_ORDER_DISTURBANCE_START = 'PHASE_ORDER_DISTURBANCE_START'
    PHASE_ORDER_DISTURBANCE_STOP = 'PHASE_ORDER_DISTURBANCE_STOP'
    RADIO_IMPACT_START = 'RADIO_IMPACT_START'
    RADIO_IMPACT_STOP = 'RADIO_IMPACT_STOP'
    DAYLIGHT_TIME_SWITCH = 'DAYLIGHT_TIME_SWITCH'
    DAYLIGHT_TIME_MODE_DATES_CHANGE = 'DAYLIGHT_TIME_MODE_DATES_CHANGE'
    INTERNAL_CLOCK_SYNC = 'INTERNAL_CLOCK_SYNC'
    METROLOGY_CHANGE = 'METROLOGY_CHANGE'
    PROFILE_CONF_CHANGE = 'PROFILE_CONF_CHANGE'
    TARIFFICATION_METHOD_CHANGE = 'TARIFFICATION_METHOD_CHANGE'
    PERMISSION_CHANGE_SETTINGS_POWER_CONTROL = 'PERMISSION_CHANGE_SETTINGS_POWER_CONTROL'
    CONTROL_LEVEL_MAINS_CHANGE = 'CONTROL_LEVEL_MAINS_CHANGE'
    PERMISSION_CHANGE_SETTINGS_CONSUMPTION_CONTROL = 'PERMISSION_CHANGE_SETTINGS_CONSUMPTION_CONTROL'
    LOAD_RELAY_CONDITION_SETTINGS_CHANGE = 'LOAD_RELAY_CONDITION_SETTINGS_CHANGE'
    SIGNALIZATION_RELAY_CONDITION_SETTINGS_CHANGE = 'SIGNALIZATION_RELAY_CONDITION_SETTINGS_CHANGE'
    INTERFACE_SIGNALIZATION_CONDITION_SETTINGS_CHANGE = 'INTERFACE_SIGNALIZATION_CONDITION_SETTINGS_CHANGE'
    INDICATION_SETTINGS_CHANGE = 'INDICATION_SETTINGS_CHANGE'
    SOUND_SIGNAL_CONDITION_SETTINGS_CHANGE = 'SOUND_SIGNAL_CONDITION_SETTINGS_CHANGE'
    LOAD_RELAY_STATE_CHANGE = 'LOAD_RELAY_STATE_CHANGE'
    SIGNALIZATION_RELAY_STATE_CHANGE = 'SIGNALIZATION_RELAY_STATE_CHANGE'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class IntegrationV0MessageMetaBSChannelProtocol(Enum):
    nbfi = 'nbfi'
    unbp = 'unbp'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class IntervalSelectValue(Enum):
    half_hour = '30 minute'
    hour = '60 minutes'
    day = '1 day'
    week = '1 week'
    month = '1 month'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class DeviceHack(Enum):
    """
    electricity_profile_packets_days_ago_is_zero: change profile window logic when days_ago = 0 when days_ago=0 fill full hours profile values after packet receive

    electricity_daily_packet_overload_value: change overload value of daily packet to 838861.0, bug was fixed in mid-April 2023 firmware 147:14:17:3:1:1

    electricity_phase_packet_generation_total_enrg_overload_value: change overload value for total energy to 33554431 in generated phase packets
    electricity_phase_packet_consumption_total_enrg_overload_value: change overload value for total energy to 33554431 in consumed phase packets
    """

    electricity_profile_packets_days_ago_is_zero = 'electricity_profile_packets_days_ago_is_zero'
    electricity_phase_packet_generated_total_enrg_overload_value = 'electricity_phase_packet_generated_total_enrg_overload_value'
    electricity_phase_packet_consumed_total_enrg_overload_value = 'electricity_phase_packet_consumed_total_enrg_overload_value'
    electricity_daily_packet_overload_value = 'electricity_daily_packet_overload_value'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class DeviceTimeTransition(Enum):
    """
    Enumeration of seasons change
    """
    summer = 'summer'
    winter = 'winter'
    unknown = 'unknown'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class DeviceClockOutOfSyncType(Enum):
    """
    Enumeration of clock sync states
    """
    synced = 'synced'
    out_of_sync_warning = 'out_of_sync_warning'
    out_of_sync_critical = 'out_of_sync_critical'
    unsynchronized = 'unsynchronized'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class DownlinkTaskType(Enum):
    time_sync = 'time_sync'
    firmware_update = 'firmware_update'
    unbp_message = 'unbp_message'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class ScheduleType(Enum):
    on_capture = 'on_capture'
    schedule = 'schedule'
    asap = 'asap'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class ReglamentType(Enum):
    ul_pr_dl_none = 'ul_pr_dl_none'
    ul_pr_dl_rm_on_capture = 'ul_pr_dl_rm_on_capture'
    ul_pr_dl_rm = 'ul_pr_dl_rm'
    ul_rm_dl_rm = 'ul_rm_dl_rm'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class SignalModulation(Enum):
    fsk = 'fsk'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class DownlinkTaskStatus(Enum):
    created = 'created'
    bs_ack = 'bs_ack'
    bs_executed_succeed = 'bs_executed_succeed'
    bs_executed_failed = 'bs_executed_failed'
    bs_skipped = 'bs_skipped'
    canceled = 'canceled'
    bs_deleted = 'bs_deleted'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class DataAggregatorApiUserType(Enum):
    base_station = "base_station"
    api_user = "api_user"
    other = "other"
    USPD = "USPD"
    universal_data_input_api = "universal_data_input_api"



from typing import Dict, Any

from data_aggregator_sdk.constants.annotations import TDevicesInfoBox


def devices_info_box_to_json(devices_info_box: TDevicesInfoBox) -> Dict[str, Any]:
    return {str(key): value for key, value in devices_info_box.items()}



from functools import wraps
from typing import Any, Callable, Dict, List, TypeVar

from ul_api_utils.errors import Client4XXInternalApiError, Server5XXInternalApiError, \
    ResponseJsonInternalApiError, NotFinishedRequestInternalApiError

from data_aggregator_sdk.errors import DataAggregatorRequestError, DataAggregatorResponseError

TKwargs = TypeVar('TKwargs', bound=Dict[str, Any])

STDResp = TypeVar('STDResp', bound=Dict[str, Any] | List[Dict[str, Any]])    # TODO:  Dict[str, Any] -> JsonApiResponsePayload

TFn = Callable[..., STDResp]


def internal_api_error_handler_old(api_method_fn: TFn) -> TFn:      # type: ignore
    ''' BROKEN TYPING '''       # TODO: must be updated to typed internal_api_error_handler
    @wraps(api_method_fn)
    def error_handler_wrapper(*args: Any, **kwargs: TKwargs) -> STDResp:    # type: ignore
        try:
            return api_method_fn(*args, **kwargs)
        except Client4XXInternalApiError as e:
            raise DataAggregatorRequestError(str(e), e, e.status_code)
        except Server5XXInternalApiError as e:
            raise DataAggregatorResponseError(str(e), e, e.status_code)
        except ResponseJsonInternalApiError as e:
            raise DataAggregatorResponseError(str(e), e, 500)
        except NotFinishedRequestInternalApiError as e:
            raise DataAggregatorResponseError("SERVICE TEMPORARY UNAVAIBLE", e, 503)
    return error_handler_wrapper



from datetime import timedelta

TIME_DELTA_TO_TZ_NAME_MAP = {
    timedelta(minutes=60 * 12.0): 'Etc/GMT+12',
    timedelta(minutes=60 * 11.0): 'Etc/GMT+11',
    timedelta(minutes=60 * 10): 'Etc/GMT+10',
    timedelta(minutes=60 * 9.5): 'Pacific/Marquesas',
    timedelta(minutes=60 * 9.0): 'Etc/GMT+9',
    timedelta(minutes=60 * 8.0): 'Etc/GMT+8',
    timedelta(minutes=60 * 7.0): 'Etc/GMT+7',
    timedelta(minutes=60 * 6.0): 'Etc/GMT+6',
    timedelta(minutes=60 * 5.0): 'Etc/GMT+5',
    timedelta(minutes=60 * 4.0): 'Etc/GMT+4',
    timedelta(minutes=60 * 3.0): 'Etc/GMT+3',
    timedelta(minutes=60 * 2.5): 'Canada/Newfoundland',
    timedelta(minutes=60 * 2.0): 'Etc/GMT+2',
    timedelta(minutes=60 * 1.0): 'Etc/GMT+1',
    timedelta(minutes=60 * -0.0): 'UTC',
    timedelta(minutes=60 * -1.0): 'Etc/GMT-1',
    timedelta(minutes=60 * -2.0): 'Etc/GMT-2',
    timedelta(minutes=60 * -3.0): 'Etc/GMT-3',
    timedelta(minutes=60 * -3.5): 'Asia/Tehran',
    timedelta(minutes=60 * -4.0): 'Etc/GMT-4',
    timedelta(minutes=60 * -4.5): 'Asia/Kabul',
    timedelta(minutes=60 * -5.0): 'Etc/GMT-5',
    timedelta(minutes=60 * -5.5): 'Asia/Colombo',
    timedelta(minutes=60 * -5.75): 'Asia/Katmandu',
    timedelta(minutes=60 * -6.0): 'Etc/GMT-6',
    timedelta(minutes=60 * -6.5): 'Asia/Yangon',
    timedelta(minutes=60 * -7.0): 'Etc/GMT-7',
    timedelta(minutes=60 * -8.0): 'Etc/GMT-8',
    timedelta(minutes=60 * -8.75): 'Australia/Eucla',
    timedelta(minutes=60 * -9.0): 'Etc/GMT-9',
    timedelta(minutes=60 * -9.5): 'Australia/North',
    timedelta(minutes=60 * -10.0): 'Etc/GMT-10',
    timedelta(minutes=60 * -10.5): 'Australia/South',
    timedelta(minutes=60 * -11.0): 'Etc/GMT-11',
    timedelta(minutes=60 * -12.0): 'Etc/GMT-12',
    timedelta(minutes=60 * -13.0): 'Etc/GMT-13',
    timedelta(minutes=60 * -13.75): 'Pacific/Chatham',
    timedelta(minutes=60 * -14.00): 'Etc/GMT-14',
}

TZ_NAME_TO_TIME_DELTA_MAP = {
    'Etc/GMT+12': timedelta(minutes=60 * 12.0),
    'Etc/GMT+11': timedelta(minutes=60 * 11.0),
    'Etc/GMT+10': timedelta(minutes=60 * 10),
    'Pacific/Marquesas': timedelta(minutes=60 * 9.5),
    'Etc/GMT+9': timedelta(minutes=60 * 9.0),
    'Etc/GMT+8': timedelta(minutes=60 * 8.0),
    'Etc/GMT+7': timedelta(minutes=60 * 7.0),
    'Etc/GMT+6': timedelta(minutes=60 * 6.0),
    'Etc/GMT+5': timedelta(minutes=60 * 5.0),
    'Etc/GMT+4': timedelta(minutes=60 * 4.0),
    'Etc/GMT+3': timedelta(minutes=60 * 3.0),
    'Canada/Newfoundland': timedelta(minutes=60 * 2.5),
    'Etc/GMT+2': timedelta(minutes=60 * 2.0),
    'Etc/GMT+1': timedelta(minutes=60 * 1.0),
    'UTC': timedelta(minutes=60 * -0.0),
    'Etc/GMT-1': timedelta(minutes=60 * -1.0),
    'Etc/GMT-2': timedelta(minutes=60 * -2.0),
    'Etc/GMT-3': timedelta(minutes=60 * -3.0),
    'Asia/Tehran': timedelta(minutes=60 * -3.5),
    'Etc/GMT-4': timedelta(minutes=60 * -4.0),
    'Asia/Kabul': timedelta(minutes=60 * -4.5),
    'Etc/GMT-5': timedelta(minutes=60 * -5.0),
    'Asia/Colombo': timedelta(minutes=60 * -5.5),
    'Asia/Katmandu': timedelta(minutes=60 * -5.75),
    'Etc/GMT-6': timedelta(minutes=60 * -6.0),
    'Asia/Yangon': timedelta(minutes=60 * -6.5),
    'Etc/GMT-7': timedelta(minutes=60 * -7.0),
    'Etc/GMT-8': timedelta(minutes=60 * -8.0),
    'Australia/Eucla': timedelta(minutes=60 * -8.75),
    'Etc/GMT-9': timedelta(minutes=60 * -9.0),
    'Australia/North': timedelta(minutes=60 * -9.5),
    'Etc/GMT-10': timedelta(minutes=60 * -10.0),
    'Australia/South': timedelta(minutes=60 * -10.5),
    'Etc/GMT-11': timedelta(minutes=60 * -11.0),
    'Etc/GMT-12': timedelta(minutes=60 * -12.0),
    'Etc/GMT-13': timedelta(minutes=60 * -13.0),
    'Pacific/Chatham': timedelta(minutes=60 * -13.75),
    'Etc/GMT-14': timedelta(minutes=60 * -14.00),
}



from data_aggregator_sdk.data_aggregator_api_sdk import DataAggregatorApiSdk
from data_aggregator_sdk.constants.enums import NetworkTypeEnum
from data_aggregator_sdk.data_aggregator_api_sdk_config import DataAggregatorApiSdkConfig
from data_aggregator_sdk.data_aggregator_sdk import DataAggregatorSdk
from data_aggregator_sdk.data_aggregator_sdk_config import DataAggregatorSdkConfig

__all__ = (
    "DataAggregatorSdk",
    "DataAggregatorSdkConfig",
    'NetworkTypeEnum',
    "DataAggregatorApiSdk",
    "DataAggregatorApiSdkConfig",
)



from datetime import datetime, timedelta


def round_dt(dt: datetime, round_delta: timedelta) -> datetime:
    dt_timezone = dt.tzinfo
    dt = dt.replace(tzinfo=None)
    rounded_dt = datetime.min + round((dt - datetime.min) / round_delta) * round_delta
    return rounded_dt.replace(tzinfo=dt_timezone)



from datetime import datetime
from typing import Any, Dict, List, Optional
from uuid import UUID

from ul_api_utils.api_resource.api_response_payload_alias import ApiBaseUserModelPayloadResponse
from ul_api_utils.api_resource.api_response import JsonApiResponsePayload

from data_aggregator_sdk.constants.enums import DeviceModificationTypeEnum, NetworkSysTypeEnum, NetworkTypeEnum, \
    ProtocolEnum, ResourceKind


class ApiDataGatewayResponse(ApiBaseUserModelPayloadResponse):
    name: str


class ApiDataGatewaysNetworkResponse(ApiBaseUserModelPayloadResponse):
    name: str
    type_network: NetworkTypeEnum
    data_gateway_id: UUID
    data_gateway: ApiDataGatewayResponse
    sys_type: NetworkSysTypeEnum
    specifier: Optional[str]
    params: Optional[Dict[str, Any]]


class ApiProtocolResponse(JsonApiResponsePayload):
    id: UUID
    date_created: datetime
    date_modified: datetime
    name: str
    type: ProtocolEnum


class ApiDataGatewayNetworkDevicePayloadResponse(JsonApiResponsePayload):
    id: UUID
    date_created: datetime
    date_modified: datetime
    uplink_protocol_id: UUID
    downlink_protocol_id: UUID
    data_gateway_network_id: UUID
    mac: int
    key_id: Optional[UUID]
    device_id: UUID
    uplink_encryption_key: Optional[str]
    downlink_encryption_key: Optional[str]
    encryption_key: Optional[str]
    protocol: ApiProtocolResponse
    network: Optional[ApiDataGatewaysNetworkResponse]


class ApiDeviceMeteringTypeResponse(JsonApiResponsePayload):
    id: UUID
    date_created: datetime
    date_modified: datetime

    sys_name: str
    name_ru: str
    name_en: str


class ApiDeviceModificationTypeResponse(JsonApiResponsePayload):
    id: UUID
    date_created: datetime
    date_modified: datetime

    sys_name: str
    name_ru: str
    name_en: str
    type: DeviceModificationTypeEnum
    metering_type_id: UUID
    device_metering_type: ApiDeviceMeteringTypeResponse


class ApiDeviceModificationResponse(JsonApiResponsePayload):
    id: UUID
    date_created: datetime
    date_modified: datetime
    name: Optional[str]
    device_modification_type_id: Optional[UUID]
    device_modification_type: ApiDeviceModificationTypeResponse


class ApiDeviceMeterPayloadResponse(ApiBaseUserModelPayloadResponse):
    device_channel_id: UUID
    value_multiplier: Optional[float]
    unit_multiplier: Optional[float]
    kind: ResourceKind


class ApiDeviceChannelPayloadResponse(ApiBaseUserModelPayloadResponse):
    inactivity_limit: Optional[int]
    serial_number: int
    device_meter: List[ApiDeviceMeterPayloadResponse]


class ApiDeviceManufacturerResponse(JsonApiResponsePayload):
    id: UUID
    date_created: datetime
    date_modified: datetime

    name: str


class ApiImportDeviceResponse(ApiBaseUserModelPayloadResponse):
    manufacturer_serial_number: str
    firmware_version: Optional[str]
    hardware_version: Optional[str]
    date_produced: Optional[datetime]
    device_manufacturer: ApiDeviceManufacturerResponse
    device_modification: ApiDeviceModificationResponse
    device_channel: List[ApiDeviceChannelPayloadResponse]
    data_gateway_network_device: Optional[ApiDataGatewayNetworkDevicePayloadResponse]



import base64
from datetime import date, datetime, time
from typing import Dict, Any, List, Tuple, Optional, Union
from uuid import UUID

from ul_api_utils.api_resource.api_response import JsonApiResponsePayload
from ul_api_utils.api_resource.api_response_payload_alias import ApiBaseUserModelPayloadResponse, ApiBaseModelPayloadResponse
from ul_api_utils.internal_api.internal_api import InternalApi
from ul_api_utils.validators.custom_fields import WhiteSpaceStrippedStr, PgTypeInt32, CronScheduleStr
from pydantic import BaseModel, UUID4, Field
from pydantic.types import NonNegativeInt

from data_aggregator_sdk.constants.enums import IntervalSelectValue, EncryptionType, ResourceKind, NetworkTypeEnum, \
    NetworkSysTypeEnum, DeviceHack, DataAggregatorApiUserType, DownlinkTaskType, ScheduleType, ReglamentType, \
    SignalModulation, DownlinkTaskStatus
from data_aggregator_sdk.data_aggregator_api_sdk_config import DataAggregatorApiSdkConfig
from data_aggregator_sdk.integration_message import ProfileKind, ProfileGranulation
from data_aggregator_sdk.types.device import ApiImportDeviceResponse, ApiDataGatewayNetworkDevicePayloadResponse, \
    ApiDeviceMeterPayloadResponse, ApiDeviceChannelPayloadResponse
from data_aggregator_sdk.types.get_data_gateway_network_device_list import ApiDataGatewaysNetworkResponse, \
    ApiDeviceModificationTypeResponse, ApiProtocolResponse
from data_aggregator_sdk.utils.internal_api_error_handler import internal_api_error_handler
from data_aggregator_sdk.utils.internal_api_error_handler_old import internal_api_error_handler_old


class ApiDeviceProfileGranularityResponse(JsonApiResponsePayload):
    device_id: UUID4 = Field(
        ...,
        title="Device ID",
        description="Each device has got unique identifier which is UUID (ex.'f273cbef-0182-4f10-a2b9-17bf54223925')",
    )
    start_date: datetime = Field(
        ...,
        title="Start date of device profile",
        description="Each device profile has start date of collecting data by time granularity",
    )
    end_date: datetime = Field(
        ...,
        title="End date of device profile",
        description="Each device profile has end date of collecting data by time granularity",
    )
    profile_kind: ProfileKind = Field(
        ...,
        title="Type of profile",
        description="Profile type is like side view of collecting data for device (ex. 'VOLTAGE_ABC', 'ENERGY_A_N')",
    )
    granularity_s: ProfileGranulation = Field(
        ...,
        title="Device profile granulation by minutes or seconds",
        description="Granulation is detailing of collecting data, resolution of graph by device (ex. 1 hour (MINUTE_60), 1 second (SECONDS_01))",
    )


class ApiDeviceProfileResponse(JsonApiResponsePayload):
    date_start: datetime = Field(
        ...,
        title="Start date of device profile",
        description="Each device profile has start date of collecting data by time granularity",
    )
    date_end: datetime = Field(
        ...,
        title="End date of device profile",
        description="Each device profile has end date of collecting data by time granularity",
    )
    values_count: int = Field(
        ...,
        title="Amount of collected values",
        description="Number of collected values in sequence of collected values for device by profile type",
    )
    values: List[Optional[float]] = Field(
        ...,
        title="Sequence of collected values",
        description="Collected values for device by profile type looks like sequence of numbers with floating point which depends on profile type",
    )
    profile_kind: ProfileKind = Field(
        ...,
        title="Type of profile",
        description="Profile type is like side view of collecting data for device (ex. 'VOLTAGE_ABC', 'ENERGY_A_N')",
    )
    granularity_s: ProfileGranulation = Field(
        ...,
        title="Device profile granulation by minutes or seconds",
        description="Granulation is detailing of collecting data, resolution of graph by device (ex. 1 hour (MINUTE_60), 1 second (SECONDS_01))",
    )


class DeviceValueDescriptionModelBase(BaseModel):
    registration_date: Optional[datetime] = Field(None, title='Registration Date', description='registration date')
    kind: ResourceKind = Field(ResourceKind.COMMON_CONSUMED, title='Kind', description=f'value kind ({", ".join(kind.name for kind in ResourceKind)})')
    tariff_number: int = Field(-1, title='Tariff Number', description='tariff number')


class DeviceValueDescriptionModel(DeviceValueDescriptionModelBase):
    device_id: Optional[UUID] = Field(None, title='Device Id', description='device identifier')


class DeviceChannelValueDescriptionModel(DeviceValueDescriptionModelBase):
    device_channel_id: Optional[UUID] = Field(None, title='Device Channel Id', description='device channel identifier')


class DeviceMeterValueDescriptionModel(DeviceValueDescriptionModelBase):
    device_meter_id: Optional[UUID] = Field(None, title='Device Meter Id', description='device meter identifier')


class ApiDeviceLastValueDateResponse(JsonApiResponsePayload):
    last_value_date: Dict[str, datetime] = Field(title='Last Value Date', description='last value date: key - device identifier, value - datetime')    # str - UUID (Device.id)


class ApiDeviceChannelLastValueDateResponse(JsonApiResponsePayload):
    last_value_date: Dict[str, datetime] = Field(title='Last Value Date', description='last value date: key - device meter identifier, value - datetime')    # str - UUID (DeviceMeter.id)


class ApiDeviceMeterLastValueDateResponse(JsonApiResponsePayload):
    last_value_date: Dict[UUID, datetime] = Field(title='Last Value Date', description='last value date: key - device meter identifier, value - datetime')   # str - UUID (DeviceMeter.id)


class GeneralApiDeviceMeter(BaseModel):
    value_multiplier: float = Field(title='Value Multiplier', description='value multiplier')
    unit_multiplier: float = Field(title='Unit Multiplier', description='unit multiplier')
    kind: ResourceKind = Field(title='Kind', description=f'meter kind ({", ".join(kind.name for kind in ResourceKind)})')


class GeneralApiDeviceChannel(BaseModel):
    serial_number: int = Field(title='Serial Number', description='Serial Number')
    inactivity_limit: Optional[int] = Field(None, title='Inactivity Limit', description='Inactivity Limit')
    device_meter: List[GeneralApiDeviceMeter] = Field(title='GeneralApiDeviceMeters', description='array of GeneralApiDeviceMeter')


class ApiDeviceManufacturerResponse(JsonApiResponsePayload):
    id: UUID4 = Field(title='Id', description='manufacturer identifier')
    date_created: datetime = Field(title='Date Created', description='date created')
    date_modified: datetime = Field(title='Date Modified', description='date modified')
    name: str = Field(title='Name', description='manufacturer name')


class ApiDeviceModificationResponse(JsonApiResponsePayload):
    id: UUID4 = Field(title='Id', description='modification identifier')
    date_created: datetime = Field(title='Date Created', description='date created')
    date_modified: datetime = Field(title='Date Modified', description='date modified')
    name: Optional[str] = Field(None, title='Name', description='modification name')
    device_modification_type_id: Optional[UUID4] = Field(None, title='Device Modification Type Id', description='device modification type identifier')
    device_modification_type: ApiDeviceModificationTypeResponse = Field(title='ApiDeviceModificationTypeResponse', description='device modification type')


class ApiDataGatewayNetworkDeviceResponse(ApiBaseUserModelPayloadResponse):
    manufacturer_serial_number: str = Field(title='Manufacturer Serial Number', description='manufacturer serial number')
    firmware_version: Optional[str] = Field(None, title='Firmware Version', description='firmware version')
    hardware_version: Optional[str] = Field(None, title='Hardware Version', description='hardware version')
    date_produced: Optional[datetime] = Field(None, title='Date Produced', description='date produced')
    device_manufacturer: ApiDeviceManufacturerResponse = Field(title='ApiDeviceManufacturerResponse', description='device manufacturer')
    device_modification: ApiDeviceModificationResponse = Field(title='ApiDeviceModificationResponse', description='device modification')
    device_channel: List[ApiDeviceChannelPayloadResponse] = Field(title='ApiDeviceChannelPayloadResponse', description='device channel')


class ApiDataGatewayNetworkDeviceByMacAndNetworkPayloadResponse(JsonApiResponsePayload):
    id: UUID4 = Field(title='Id', description='ID')
    date_created: datetime = Field(title='Date Created', description='date created')
    date_modified: datetime = Field(title='Date Modified', description='date modified')
    uplink_protocol_id: UUID4 = Field(title='Uplink Protocol Id', description='uplink protocol identifier')
    downlink_protocol_id: UUID4 = Field(title='Downlink Protocol Id', description='downlink protocol identifier')
    data_gateway_network_id: UUID4 = Field(title='Data Gateway Network Id', description='data gateway network identifier')
    mac: int = Field(title='Mac', description='MAC')
    key_id: Optional[UUID4] = Field(None, title='Key Id', description='key id')
    device_id: UUID4 = Field(title='Device Id', description='device id')
    device: ApiDataGatewayNetworkDeviceResponse = Field(title='ApiDataGatewayNetworkDeviceResponse', description='device')
    uplink_encryption_key: Optional[str] = Field(None, title='Uplink Encryption Key', description='uplink encryption key')
    downlink_encryption_key: Optional[str] = Field(None, title='Downlink Encryption Key', description='downlink encryption key')
    encryption_key: Optional[str] = Field(None, title='Encryption Key', description='encryption key')
    protocol: ApiProtocolResponse = Field(title='ApiProtocolResponse', description='protocol')
    network: Optional[ApiDataGatewaysNetworkResponse] = Field(None, title='ApiDataGatewaysNetworkResponse', description='data gateway network')


class DeviceDescriptionModel(BaseModel):
    device_id: Union[UUID4, UUID] = Field(title='Device Id', description='device id')
    serial_number: int = Field(title='Serial Number', description='serial number')
    registration_date: Optional[datetime] = Field(None, title='Registration Date', description='registration date')
    kind: ResourceKind = Field(ResourceKind.COMMON_CONSUMED, title='Kind', description=f'meter kind ({", ".join(kind.name for kind in ResourceKind)})')
    tariff_number: int = Field(-1, title='Tariff Number', description='tariff number')


class GeneralDeviceValueDescriptionModel(BaseModel):
    device_id: Union[UUID4, UUID] = Field(title='Device Id', description='device id')
    serial_number: int = Field(title='Serial Number', description='serial number')
    registration_date: Optional[datetime] = Field(None, title='Registration Date', description='registration date')
    kind: ResourceKind = Field(ResourceKind.COMMON_CONSUMED, title='Kind', description=f'meter kind ({", ".join(kind.name for kind in ResourceKind)})')
    tariff_number: int = Field(-1, title='Tariff Number', description='tariff number')


class GeneralDeviceMeterValueDescriptionModel(BaseModel):
    meter_id: UUID = Field(title='Meter Id', description='device meter identifier')
    registration_date: Optional[datetime] = Field(None, title='Registration Date', description='registration date')
    kind: ResourceKind = Field(ResourceKind.COMMON_CONSUMED, title='Kind', description=f'meter kind ({", ".join(kind.name for kind in ResourceKind)})')
    tariff_number: int = Field(-1, title='Tariff Number', description='tariff number')


class DeviceBatteryLvlDescriptionModel(BaseModel):
    device_id: Union[UUID4, UUID] = Field(title='Device Id', description='device id')
    registration_date: Optional[datetime] = Field(None, title='Registration Date', description='registration date')


class GeneralDeviceValueModel(BaseModel):
    devices: List[GeneralDeviceValueDescriptionModel] = Field(title='GeneralDeviceValueDescriptionModels', description='array of GeneralDeviceValueDescriptionModel')


class GeneralDeviceMeterValueModel(BaseModel):
    devices_meter: List[GeneralDeviceMeterValueDescriptionModel] = Field(title='GeneralDeviceValueDescriptionModels', description='array of GeneralDeviceValueDescriptionModel')


class GeneralDeviceModificationModel(BaseModel):
    devices: List[DeviceDescriptionModel] = Field(title='DeviceDescriptionModel', description='array of DeviceDescriptionModel')


class DeviceValueBatteryLvlModel(BaseModel):
    devices: List[DeviceBatteryLvlDescriptionModel] = Field(title='DeviceBatteryLvlDescriptionModels', description='array of DeviceBatteryLvlDescriptionModel')


class GeneralApiDeviceResponse(BaseModel):
    date: datetime = Field(title='Date', description='Date')
    device_id: UUID = Field(title='Device Id', description='device identifier')
    serial_number: int = Field(title='Serial Number', description='serial number')
    value: float = Field(title='Value', description='Value')
    kind: ResourceKind = Field(title='Kind', description=f'Kind ({", ".join(kind.name for kind in ResourceKind)})')
    tariff_number: int = Field(title='Tariff Number', description='tariff number')


class GeneralApiDeviceMeterResponse(BaseModel):
    meter_id: UUID = Field(title='Meter Id', description='meter identifier')
    value: float = Field(title='Value', description='Value')
    value_date: datetime = Field(title='Value Date', description='Value Date')
    kind: ResourceKind = Field(title='Kind', description=f'Kind ({", ".join(kind.name for kind in ResourceKind)})')
    tariff_number: int = Field(title='Tariff Number', description='tariff number')


class GeneralApiDeviceValueResponse(BaseModel):
    device_id: UUID = Field(title='Device Id', description='device identifier')
    serial_number: int = Field(title='Serial Number', description='serial number')
    kind: ResourceKind = Field(title='Kind', description=f'Kind ({", ".join(kind.name for kind in ResourceKind)})')
    tariff_number: int = Field(title='Tariff Number', description='tariff number')
    value: float = Field(title='Value', description='Value')
    value_date: datetime = Field(title='Value Date', description='Value Date')
    last_value: Optional[float] = Field(None, title='Last Value', description='last value')
    last_value_date: Optional[datetime] = Field(None, title='Last Value Date', description='last value date')


class GeneralApiDeviceMeterValueResponse(JsonApiResponsePayload):
    meter_id: UUID = Field(title='Meter Id', description='meter identifier')
    kind: ResourceKind = Field(title='Kind', description=f'Kind ({", ".join(kind.name for kind in ResourceKind)})')
    tariff_number: int = Field(title='Tariff Number', description='tariff number')
    value: float = Field(title='Value', description='Value')
    value_date: datetime = Field(title='Value Date', description='Value Date')
    last_value: Optional[float] = Field(None, title='Last Value', description='last value')
    last_value_date: Optional[datetime] = Field(None, title='Last Value Date', description='last value date')


class ApiDeviceLowBatteryResponse(JsonApiResponsePayload):
    value_date: datetime = Field(title='Value Date', description='Value Date')
    device_id: UUID = Field(title='Device Id', description='device identifier')
    value: float = Field(title='Value', description='Value')
    last_value: Optional[float] = Field(None, title='Last Value', description='last value')
    last_value_date: Optional[datetime] = Field(None, title='Last Value Date', description='last value date')


class ApiDeviceResponse(ApiBaseUserModelPayloadResponse):
    manufacturer_serial_number: str = Field(title='Manufacturer Serial Number', description='manufacturer serial number')
    firmware_version: Optional[str] = Field(None, title='Firmware Version', description='firmware version')
    hardware_version: Optional[str] = Field(None, title='Hardware Version', description='hardware version')
    date_produced: Optional[datetime] = Field(None, title='Date Produced', description='date produced')
    device_manufacturer: ApiDeviceManufacturerResponse = Field(title='Device Manufacturer', description='device manufacturer')
    device_modification: ApiDeviceModificationResponse = Field(title='Device Modification', description='device modification')
    device_channel: List[ApiDeviceChannelPayloadResponse] = Field(title='ApiDeviceChannelPayloadResponse', description='array of ApiDeviceChannelPayloadResponse')
    data_gateway_network_device: Optional[ApiDataGatewayNetworkDevicePayloadResponse] = Field(None, title='ApiDataGatewayNetworkDevicePayloadResponses', description='array of ApiDataGatewayNetworkDevicePayloadResponse')


class ApiDeviceChannelResponse(BaseModel):
    response: Dict[str, Any] = Field(title='Response', description='response')


class ApiDeviceNoData(JsonApiResponsePayload):
    device_id: UUID4 = Field(title='Device Id')
    serial_number: int = Field(title='Serial Number')
    kind: ResourceKind = Field(title='Kind', description=f'Kind ({", ".join(kind.name for kind in ResourceKind)})')
    tariff_number: int = Field(title='Tariff Number')
    last_value: Optional[float] = Field(None, title='Last Value')
    last_value_date: Optional[date] = Field(title='Last Value Date')
    count_days: int = Field(title='Count Days', description='count dates')


class ApiDeviceNoDataResponse(JsonApiResponsePayload):
    objects: List[ApiDeviceNoData] = Field(title='ApiDeviceNoDatas', description='array of ApiDeviceNoData')
    procent_no_data: float = Field(title='Procent No Data', description='percent no data')
    count_devices: int = Field(title='Count Devices', description='count devices')
    total_count: int = Field(title='Total Count', description='total count')


class ApiDeviceManufacturer(BaseModel):
    manufacturer_serial_number: str = Field(title='Manufacturer Serial Number', description='manufacturer serial number')


class ApiDeviceFactoryParameters(BaseModel):
    manufacturer_id: Optional[UUID4] = Field(None, title='Manufacturer Id', description='manufacturer id')
    protocol_id: Optional[UUID4] = Field(None, title='Protocol Id', description='protocol id')
    date_produced: Optional[datetime] = Field(None, title='Date Produced', description='date produced')
    device_modification_type_id: Optional[UUID4] = Field(None, title='Device Modification Type Id', description='device modification type id')


class ApiDeviceValues(BaseModel):
    devices: List[List[Any]] = Field(title='Devices', description='devices')


class ApiValueDeviceChannelForEripReports(BaseModel):
    devices: List[Tuple[UUID4, int]] = Field(title='Devices', description='array tuple device identifier, channel number')
    date_to: datetime = Field(title='Date To', description='date to')


class ApiDeviceEvent(BaseModel):
    devices: List[UUID4] = Field(title='Devices', description='array of device identifiers')


class ApiDeviceNoDataList(BaseModel):
    devices: List[DeviceDescriptionModel] = Field(title='DeviceDescriptionModel', description='array of DeviceDescriptionModel')
    period_from: datetime = Field(title='Period From', description='period from (datetime)')
    period_to: Optional[datetime] = Field(None, title='Period To', description='period to (datetime)')


class ApiDataGatewayNetworkDevice(JsonApiResponsePayload):
    id: UUID4 = Field(title='Id', description='identifier')
    date_created: datetime = Field(title='Date Created', description='date created')
    date_modified: datetime = Field(title='Date Modified', description='date modified')
    uplink_protocol_id: UUID4 = Field(title='Uplink Protocol Id', description='uplink protocol identifier')
    downlink_protocol_id: UUID4 = Field(title='Downlink Protocol Id', description='downlink protocol identifier')
    mac: int = Field(title='Mac', description='MAC')
    key_id: Optional[UUID4] = Field(None, title='Key Id', description='key identifier')
    device_id: UUID4 = Field(title='Device Id', description='device id')
    uplink_encryption_key: Optional[str] = Field(None, title='Uplink Encryption Key', description='decryption key')
    downlink_encryption_key: Optional[str] = Field(None, title='Downlink Encryption Key', description='encryption key')
    uplink_encryption_type: EncryptionType = Field(title='Uplink Encryption Type', description=f'uplink encryption type ({", ".join(encryption.name for encryption in EncryptionType)})')
    downlink_encryption_type: EncryptionType = Field(title='Downlink Encryption Type', description=f'downlink encryption type ({", ".join(encryption.name for encryption in EncryptionType)})')
    protocol: ApiProtocolResponse = Field(title='ApiProtocolResponse', description='protocol')
    network: ApiDataGatewaysNetworkResponse = Field(title='ApiDataGatewaysNetworkResponse', description='DataGatewayNetwork')

    @property
    def downlink_encryption_key_bytes(self) -> bytes:
        downlink_encryption_key = self.downlink_encryption_key if self.downlink_encryption_key is not None else ''
        return base64.b64decode(downlink_encryption_key)

    @property
    def uplink_encryption_key_bytes(self) -> bytes:
        uplink_encryption_key = self.uplink_encryption_key if self.uplink_encryption_key is not None else ''
        return base64.b64decode(uplink_encryption_key)


class DeviceData(JsonApiResponsePayload):
    manufacturer_serial_number: str = Field(title='Manufacturer Serial Number', description='device manufacturer serial number')
    firmware_version: Optional[str] = Field(None, title='Firmware Version', description='device firmware version')
    hardware_version: Optional[str] = Field(None, title='Hardware Version', description='device hardware version')
    hacks: List[DeviceHack] = Field(title='Hacks', description=f'array of device processing modifiers ({", ".join(hack.name for hack in DeviceHack)})')
    device_tz: Optional[str] = Field(None, title='Device Tz', description='device time zone')
    date_produced: Optional[datetime] = Field(None, title='Date Produced', description='date produced')
    device_manufacturer: ApiDeviceManufacturerResponse = Field(title='ApiDeviceManufacturerResponse', description='device manufacturer')
    device_modification: ApiDeviceModificationResponse = Field(title='ApiDeviceModificationResponse', description='device modification')
    device_channel: List[ApiDeviceChannelPayloadResponse] = Field(title='ApiDeviceChannelPayloadResponses', description='device channel')
    all_data_gateway_networks: List[ApiDataGatewayNetworkDevice] = Field(title='ApiDataGatewayNetworkDevice', description='all data gateway networks')


class ApiValueDeviceListResponse(JsonApiResponsePayload):
    value_date: datetime = Field(title='Value Date', description='value date')
    device_id: UUID4 = Field(title='Device Id', description='device id')
    serial_number: int = Field(title='Serial Number', description='serial number')
    kind: ResourceKind = Field(title='Kind', description=f'kind ({", ".join(kind.name for kind in ResourceKind)})')
    tariff_number: int = Field(title='Tariff Number', description='tariff number')
    value: float = Field(title='Value', description='VALUE')
    last_value: Optional[float] = Field(None, title='Last Value', description='last value')
    last_value_date: Optional[datetime] = Field(None, title='Last Value Date', description='last value date')


class ApiDeviceFirstValueDateResponse(JsonApiResponsePayload):
    date: datetime = Field(title='Date', description='date')
    device_id: UUID4 = Field(title='Device Id', description='device id')
    serial_number: int = Field(title='Serial Number', description='serial number')
    value: float = Field(title='Value', description='value')
    kind: ResourceKind = Field(title='Kind', description=f'kind ({", ".join(kind.name for kind in ResourceKind)})')
    tariff_number: int = Field(title='Tariff Number', description='tariff number')


class ApiMeterFirstValueDateResponse(JsonApiResponsePayload):
    meter_id: UUID = Field(title='Meter Id', description='device meter identifier')
    value: float = Field(title='Value', description='value')
    value_date: datetime = Field(title='Value Date', description='value date')
    kind: ResourceKind = Field(title='Kind', description=f'kind ({", ".join(kind.name for kind in ResourceKind)})')
    tariff_number: int = Field(title='Tariff Number', description='tariff number')


class ApiLastValueEripReportsResponse(JsonApiResponsePayload):
    device_id: UUID4 = Field(title='Device Id', description='device id')
    serial_number: int = Field(title='Serial Number', description='serial number')
    date: datetime = Field(title='Date', description='date')
    value: float = Field(title='Value', description='value')
    kind: ResourceKind = Field(title='Kind', description=f'kind ({", ".join(kind.name for kind in ResourceKind)})')
    tariff_number: int = Field(title='Tariff Number', description='tariff number')


class DeviceMac(JsonApiResponsePayload):
    mac: int = Field(title='Mac', description='MAC')
    device_id: UUID4 = Field(title='Device Id', description='device identifier')
    address_id: UUID4 = Field(title='Address Id', description='address identifier')


class DeviceImbalanceObject(JsonApiResponsePayload):
    object_procent: float = Field(title='Object Procent', description='object percent')
    different: float = Field(title='Different', description='different')
    flat_sum: float = Field(title='Flat Sum', description='flat sum')
    group_sum: float = Field(title='Group Sum', description='group sum')
    group_devices: List[DeviceMac] = Field(title='Group Devices', description='array of DeviceMac')
    count_not_data_flat: int = Field(title='Count Not Data Flat', description='count not data flat')
    all_flat_count: int = Field(title='All Flat Count', description='all flat count')
    address_id: UUID4 = Field(title='Address Id', description='address id')


class ApiImbalanceValueDeviceListResponse(JsonApiResponsePayload):
    procent: float = Field(title='Procent', description='percent')
    imbalance: float = Field(title='Imbalance', description='imbalance value')
    object: List[DeviceImbalanceObject] = Field(title='Object', description='array of DeviceImbalanceObject')


class SelectValueResultObject(JsonApiResponsePayload):
    number_object: Optional[str] = Field(None, title='Number Object', description='number object')
    count_flat_devices: Optional[int] = Field(None, title='Count Flat Devices', description='count flat devices')
    count_not_data_flat: Optional[int] = Field(None, title='Count Not Data Flat', description='count not data flat')
    device_id: List[UUID4] = Field(title='Device Id', description='device id')
    last_date: date = Field(title='Last Date', description='last date')
    previous_last_value: float = Field(title='Previous Last Value', description='previous last value')
    current_last_value: float = Field(title='Current Last Value', description='current last value')
    address_flat_node_id: Optional[str] = Field(title='Address Flat Node Id', description='address flat node id')
    mac: Optional[int] = Field(None, title='Mac', description='MAC')


class ApiObjectValuesDeviceListResponse(JsonApiResponsePayload):
    count_object_devices: int = Field(title='Count Object Devices', description='count object devices')
    count_not_data_objects_devices: int = Field(title='Count Not Data Objects Devices', description='count not data objects devices')
    objects: List[SelectValueResultObject] = Field(title='Objects', description='array of SelectValueResultObject')


class DeviceEventsResponse(BaseModel):
    event_id: UUID4 = Field(title='Event Id', description='device event identifier')
    date_created: str = Field(title='Date Created', description='date created')
    user_created_id: UUID4 = Field(title='User Created Id', description='user created id')
    value: Optional[float] = Field(None, title='Value', description='device event value')
    device_id: UUID4 = Field(title='Device Id', description='device identifier')
    serial_number: Optional[int] = Field(title='Serial Number', description='device channel serial number')
    type: str = Field(title='Type', description='event type')
    date: str = Field(title='Date', description='event date')
    data: Optional[Dict[str, Any]] = Field(None, title='Data', description='event extra data')
    is_system_generated: bool = Field(title='Is System Generated', description='is system generated')


class DeviceStateSyncDataResponse(JsonApiResponsePayload):
    device_events: List[DeviceEventsResponse] = Field(title='Device Events', description='array of DeviceEventsResponse')
    offset_id: Optional[UUID4] = Field(None, title='Offset Id', description='Data State Sync Offset identifier')


class UpdateSyncOffsetResponse(JsonApiResponsePayload):
    id: UUID4 = Field(title='Id', description='device state sync offset identifier')
    date_created: datetime = Field(title='Date Created', description='device state sync offset date created')
    date_modified: datetime = Field(title='Date Modified', description='device state sync offset date modified')
    user_created_id: UUID4 = Field(title='User Created Id', description='user created identifier')
    user_modified_id: UUID4 = Field(title='User Modified Id', description='user modified identifier')
    api_user_id: UUID4 = Field(title='Api User Id', description='api user identifier')
    type: str = Field(title='Type', description='device state sync offset type')
    offset: datetime = Field(title='Offset', description='device state sync offset')
    is_confirmed: bool = Field(title='Is Confirmed', description='device state sync offset is confirmed')


class ApiModificationResponse(BaseModel):
    sys_name: str = Field(title='System Name', description='system name')
    name_ru: str = Field(title='Name Ru', description='name in russian language')
    name_en: str = Field(title='Name En', description='name in english language')


class ApiModificationResponseForDevices(JsonApiResponsePayload):
    device_modification: Dict[str, ApiModificationResponse] = Field(title='Device Modification', description='key - device identifier, value - device modification')


class ApiLastValueDateResponseForDevices(JsonApiResponsePayload):
    last_value_date: Dict[str, datetime] = Field(title='Last value date', description='key - device identifier, value - date of last value')


class ApiDeviceImbalanceModel(BaseModel):
    flat_devices: Optional[List[List[Any]]] = Field(title='Flat Devices', description='flat devices')
    group_devices: Optional[List[List[Any]]] = Field(title='Group Devices', description='group devices')


class ApiBinaryDataResponse(JsonApiResponsePayload):
    id: UUID4 = Field(title='Id', description='binary data identifier')
    date_created: datetime = Field(title='Date Created', description='date created')
    user_created_id: UUID4 = Field(title='User Created Id', description='user created id')
    hash: str = Field(title='Hash', description='binary data hash')
    content: str = Field(title='Content', description='binary data content')
    file_name: str = Field(title='File Name', description='file name')


class ApiDataGatewayNetworkDeviceForBsUplinkSignalResponse(JsonApiResponsePayload):
    id: UUID4
    date_created: datetime
    date_modified: datetime
    uplink_protocol_id: UUID4
    downlink_protocol_id: UUID4
    mac: int
    key_id: Optional[UUID4]
    device_id: UUID4
    uplink_encryption_key: Optional[str]
    downlink_encryption_key: Optional[str]
    uplink_encryption_type: EncryptionType
    downlink_encryption_type: EncryptionType


class ApiBSUplinkSignalResponse(JsonApiResponsePayload):
    id: UUID4
    date_created: datetime
    user_created_id: UUID4

    rssi: Optional[int]
    snr: Optional[int]
    data_gateway_network_device_id: UUID

    data_gateway_network_device: Optional[ApiDataGatewayNetworkDeviceForBsUplinkSignalResponse]
    bs_id: UUID


class BaseBSDownLinkStatusLogData(BaseModel):
    effective_date: Optional[datetime]
    is_active: bool
    bs_downlink_task_id: UUID
    status: DownlinkTaskStatus
    status_comment: str


class ApiBSDownLinkStatusLogResponse(BaseBSDownLinkStatusLogData, JsonApiResponsePayload):
    id: UUID4
    date_created: datetime
    user_created_id: UUID4


class BaseApiBSDownlinkTask(JsonApiResponsePayload):
    note: WhiteSpaceStrippedStr
    type: DownlinkTaskType
    priority: PgTypeInt32
    broadcast: bool
    schedule_type: ScheduleType
    schedule_effective_date_from: datetime
    schedule_effective_date_to: datetime
    reglament_type: ReglamentType
    signal_power: int
    signal_freq: int
    signal_baudrate: int
    signal_modulation: SignalModulation
    lbt_enable: bool
    lbt_max_waiting_time_ms: int
    lbt_silent_time_ms: int
    force: bool
    retry_count: NonNegativeInt
    retry_min_delay_ms: int
    min_tx_delay_after_rx_ms: Optional[int]
    tx_duration_ms: Optional[int]


class ApiBSDownlinkTaskResponse(BaseApiBSDownlinkTask):
    id: UUID4
    date_created: datetime
    user_created_id: UUID4
    schedule_time: Optional[list[time]]
    schedule_cron: Optional[list[CronScheduleStr]]
    bs_downlink_task_time_sync_id: Optional[UUID4]
    bs_downlink_task_firmware_update_id: Optional[UUID4]
    bs_downlink_task_unbp_message_id: Optional[UUID4]
    current_status_log: ApiBSDownLinkStatusLogResponse


class ApiDataGatewayNetworkResponse(ApiBaseUserModelPayloadResponse):
    name: str
    type_network: NetworkTypeEnum
    data_gateway_id: UUID4
    sys_type: NetworkSysTypeEnum
    specifier: Optional[str]
    params: Optional[Dict[str, Any]]


class DataAggregatorApiUserResponse(ApiBaseModelPayloadResponse):
    date_expiration: datetime = Field(..., title='Date expiration', description='Token date expiration')
    name: str = Field(..., title='Name', description='User name')
    note: str = Field(..., title='Note', description='User notes')
    permissions: List[int] = Field(..., title='Permissions', description='User permsissions')
    type: DataAggregatorApiUserType = Field(..., title='User type', description='User type:')
    data_gateway_network_id: Optional[UUID] = Field(None, title='Data Gateway Network Id', description='Data Gateway Network Identifier')
    data_gateway_network: Optional[ApiDataGatewayNetworkResponse] = Field(None, title='Data Gateway Network', description='Data Gateway Network')


class DataAggregatorApiSdk:

    def __init__(self, config: DataAggregatorApiSdkConfig) -> None:
        self._config = config
        # self._api = InternalApi(entry_point=self._config.api_url, default_auth_token=self._config.api_token)
        self._api_device = InternalApi(entry_point=self._config.api_device_url, default_auth_token=self._config.api_token)

        # self._api_base_station: Optional[InternalApi] = None
        # if self._config.api_base_station_url is not None:
        #     self._api_base_station = InternalApi(entry_point=self._config.api_base_station_url, default_auth_token=self._config.api_token)

#     '''used in services: not used'''
#     @internal_api_error_handler
#     def get_device_dict_by_mac_and_network(self, mac: str, gateway_id: UUID, network_id: UUID) -> Dict[str, Any]:  # TODO: typing
#         return self._api_device.request_get(f'/data-gateways/{gateway_id}/networks/{network_id}/device_mac/{mac}').check().payload_raw    # type: ignore

#     '''used in services: not used'''
#     @internal_api_error_handler
#     def get_device_data(self, gateway_id: UUID, network_id: UUID, mac: int) -> DeviceData:
#         return self._api_device.request_get(f'/data-gateways/{gateway_id}/networks/{network_id}/device_mac/{mac}/data')\
#             .typed(DeviceData).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_devices_data(self, gateway_id: UUID, network_id: UUID, mac_list: List[int]) -> List[DeviceData]:
        return self._api_device.request_post(f'/data-gateways/{gateway_id}/networks/{network_id}/data', json={"mac_list": mac_list})\
            .typed(List[DeviceData]).check().payload

    '''used in services: Service Data Gateway'''
    @internal_api_error_handler
    def get_device(self, device_id: UUID) -> ApiDeviceResponse:
        return self._api_device.request_get(f'/devices/{device_id}').typed(ApiDeviceResponse).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_device_channels(self, device_id: UUID) -> List[ApiDeviceChannelPayloadResponse]:
        return self._api_device.request_get(f'/device/{device_id}/device_channel/').typed(List[ApiDeviceChannelPayloadResponse]).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_device_meters(self, device_channel_id: UUID) -> List[ApiDeviceMeterPayloadResponse]:
        return self._api_device.request_get(f'/device-channels/{device_channel_id}/device-meters').typed(List[ApiDeviceMeterPayloadResponse]).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def update_device_meter_unit_multiplier(
        self,
        device_meter_id: UUID,
        device_channel_id: UUID,
        unit_multiplier: float,
        date_application_from: Optional[date],
        date_application_to: Optional[date],
    ) -> ApiDeviceMeterPayloadResponse:
        return self._api_device.request_put(
            f'/device-channels/{device_channel_id}/device-meters/{device_meter_id}/unit-multiplier',
            json={
                "unit_multiplier": unit_multiplier,
                "date_application_from": date_application_from,
                "date_application_to": date_application_to,
            },
        ).typed(ApiDeviceMeterPayloadResponse).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_or_create_channel(
        self,
        device_id: UUID,
        serial_number: int,
        inactivity_limit: Optional[int],
    ) -> ApiDeviceChannelPayloadResponse:
        return self._api_device.request_post('/device_channel', json={
            "device_id": str(device_id),
            "serial_number": serial_number,
            "inactivity_limit": inactivity_limit,
        }).typed(ApiDeviceChannelPayloadResponse).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_or_create_meter(
        self,
        device_channel_id: UUID,
        value_multiplier: Optional[float],
        unit_multiplier: Optional[float],
    ) -> ApiDeviceMeterPayloadResponse:
        return self._api_device.request_post('/device-meter', json={
            device_channel_id: str(device_channel_id),
            value_multiplier: value_multiplier,
            unit_multiplier: unit_multiplier,
        }).typed(ApiDeviceMeterPayloadResponse).check().payload

#     '''used in services: not used'''
#     @internal_api_error_handler
#     def get_device_networks(
#         self,
#         device_id: UUID,
#         type_network: NetworkTypeEnum,
#         limit: Optional[int],
#         offset: Optional[int],
#         filters: Optional[List[Dict[str, Any]]],
#         sorts: Optional[List[Tuple[str, str]]],
#     ) -> List[Dict[str, Any]]:  # TODO: typing
#         if type_network not in NetworkTypeEnum:
#             raise ValueError("NetworkType can be only input or output")
#         return self._api_device.request_get(f'/device-network/{device_id}/type/{type_network.value}', q={
#             "filters": filters,
#             "sorts": sorts,
#             "limit": limit,
#             "offset": offset,
#         }).check().payload_raw    # type: ignore

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_device_manufacturers(
        self,
        limit: Optional[int],
        offset: Optional[int],
        filters: Optional[List[Dict[str, Any]]],
        sorts: Optional[List[Tuple[str, str]]],
    ) -> List[ApiDeviceManufacturerResponse]:
        return self._api_device.request_get(
            '/manufacturers',
            q={
                "filters": filters,
                "sorts": sorts,
                "limit": limit,
                "offset": offset,
            }).typed(List[ApiDeviceManufacturerResponse]).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_protocols(
        self,
        limit: Optional[int],
        offset: Optional[int],
        filters: Optional[List[Dict[str, Any]]],
        sorts: Optional[List[Tuple[str, str]]],
    ) -> List[ApiProtocolResponse]:
        return self._api_device.request_get(
            '/protocols',
            q={"filters": filters, "sorts": sorts, "limit": limit, "offset": offset},
        ).typed(List[ApiProtocolResponse]).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_device_modification_types(
        self,
        limit: Optional[int],
        offset: Optional[int],
        filters: Optional[List[Dict[str, Any]]],
        sorts: Optional[List[Tuple[str, str]]],
    ) -> List[ApiDeviceModificationTypeResponse]:
        return self._api_device.request_get(
            '/device-modification-types',
            q={"filters": filters, "sorts": sorts, "limit": limit, "offset": offset},
        ).typed(List[ApiDeviceModificationTypeResponse]).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_device_factory_parameters(self, devices_id: Union[UUID, List[UUID]]) -> List[ApiDeviceResponse]:
        devices = [devices_id] if not isinstance(devices_id, list) else devices_id
        return self._api_device.request_post(
            '/devices/factory-parameters',
            json={"devices": [str(device) for device in devices]},
        ).typed(List[ApiDeviceResponse]).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler_old
    def get_device_short_factory_parameters(self, device_ids: Union[UUID, List[UUID]]) -> Dict[str, Any]:  # TODO: typing
        devices = [device_ids] if not isinstance(device_ids, list) else device_ids
        return self._api_device.request_post(
            '/devices/short-factory-parameters',
            json={"devices": devices},
        ).check().payload_raw    # type: ignore

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def update_device_manufacturer_by_id(self, device_id: UUID, manufacturer_serial_number: str) -> ApiDeviceResponse:
        return self._api_device.request_patch(
            f'/devices/{device_id}/serial_number',
            json={'manufacturer_serial_number': manufacturer_serial_number},
        ).typed(ApiDeviceResponse).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def update_device_factory_parameters(
        self,
        device_id: UUID,
        manufacturer_id: Optional[UUID],
        protocol_id: Optional[UUID],
        date_produced: Optional[datetime],
        device_modification_id: Optional[UUID],
        device_modification_type_id: Optional[UUID],
    ) -> ApiDeviceResponse:
        body_data: Dict[str, Any] = {'date_produced': str(date_produced) if date_produced else None}
        if manufacturer_id is not None:
            body_data.update({'manufacturer_id': str(manufacturer_id)})

        if protocol_id is not None:
            body_data.update({'protocol_id': str(protocol_id)})

        if device_modification_id is not None:
            body_data.update({'device_modification_id': device_modification_id})

        if device_modification_type_id is not None:
            body_data.update({'device_modification_type_id': str(device_modification_type_id)})

        return self._api_device.request_patch(
            f'/device/{device_id}/factory-parameters',
            json=body_data,
        ).typed(ApiDeviceResponse).check().payload

#   '''used in services: not used'''
#     @internal_api_error_handler
#     def get_device_logger_data(
#         self,
#         device_id: UUID,
#     ) -> Dict[str, Any]:  # TODO: typing
#         return self._api_device.request_get(
#             f'/devices/{device_id}/logger_data').check().payload_raw    # type: ignore

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_value_for_device_list(
        self,
        period_from: date | datetime,
        period_to: date | datetime,
        iteration_interval: Optional[IntervalSelectValue],
        devices: GeneralDeviceValueModel,
        locf: Optional[bool],
    ) -> List[ApiValueDeviceListResponse]:
        return self._api_device.request_post('/list-device/value-for-device-list', json=devices, q={
            'period_from': period_from,
            'period_to': period_to,
            'locf': locf,
            'iteration_interval': iteration_interval.value if iteration_interval is not None else None,
        }).typed(List[ApiValueDeviceListResponse]).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_value_for_device_meter(
        self,
        period_from: date | datetime,
        period_to: date | datetime,
        iteration_interval: Optional[IntervalSelectValue],
        devices_meter: GeneralDeviceMeterValueModel,
    ) -> List[GeneralApiDeviceMeterValueResponse]:
        return self._api_device.request_post('/list-device/value-for-devices-meter', json=devices_meter, q={
            'period_from': period_from,
            'period_to': period_to,
            'iteration_interval': iteration_interval.value if iteration_interval is not None else None,
        }).typed(List[GeneralApiDeviceMeterValueResponse]).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_battery_lvl_for_device_list(
        self,
        period_from: date,
        period_to: date,
        iteration_interval: Optional[IntervalSelectValue],
        devices: DeviceValueBatteryLvlModel,
        locf: Optional[bool],
    ) -> List[ApiDeviceLowBatteryResponse]:
        return self._api_device.request_post('/list-device/battery-lvl-for-device-list', json=devices, q={
            'period_from': period_from,
            'period_to': period_to,
            'locf': locf,
            'iteration_interval': iteration_interval.value if iteration_interval is not None else None,
        }).typed(List[ApiDeviceLowBatteryResponse]).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_first_value_date_for_devices(
        self,
        devices: GeneralDeviceValueModel,
    ) -> List[ApiDeviceFirstValueDateResponse]:
        return self._api_device.request_post(
            '/list-device/search-first-value-date',
            json=devices,
        ).typed(List[ApiDeviceFirstValueDateResponse]).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_first_value_date_for_meters(
        self,
        meters: GeneralDeviceMeterValueModel,
    ) -> List[ApiMeterFirstValueDateResponse]:
        return self._api_device.request_post(
            '/meters/first-value-date',
            json=meters,
        ).typed(List[ApiMeterFirstValueDateResponse]).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_last_value_for_erip_reports(
        self,
        devices: List[Tuple[UUID4, int]],
        date_to: datetime,
    ) -> List[ApiLastValueEripReportsResponse]:
        # devices = List[List['device_id: UUID', serial_number: int]]
        return self._api_device.request_post(
            '/erip-reports/last-value',
            json=ApiValueDeviceChannelForEripReports(date_to=date_to, devices=devices),
        ).typed(List[ApiLastValueEripReportsResponse]).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_object_values_by_device_list(
        self,
        devices: Dict[str, List[List[Any]]],
        reporting_period: date,
    ) -> List[ApiObjectValuesDeviceListResponse]:
        # Dict["name_object": List[List['device_id: UUID', channel: int]],
        # "name_object": List[List['device_id: UUID', channel: int]]]
        return self._api_device.request_post(
            '/list-device/value-object-period',
            q={'reporting_period': reporting_period},
            json={'devices': devices},
        ).typed(List[ApiObjectValuesDeviceListResponse]).check().payload

#   '''used in services: not used'''
# Сделан с соответствием с новой концепцией насчет DataAggregator (get_object_values_by_device_list)
#     @internal_api_error_handler
#     def get_consumption_period_by_device_list(self, devices: List[List[Any]], reporting_period: date) -> List[Dict[str, Any]]:  # TODO: typing
#         # List[List['device_id: UUID', channel: int]]
#         return self._api.request_post('/list-device/consumption-period', q={'reporting_period': reporting_period}, json={'devices': devices}).check().payload_raw

#   '''used in services: not used'''
#     @internal_api_error_handler
#     def get_object_delta(self, devices: Dict[str, List[List[Any]]]) -> List[Dict[str, Any]]:  # TODO: typing
#         # Dict["name_object": List[List['device_id: UUID', channel: int]],
#         # "name_object": List[List['device_id: UUID', channel: int]]]
#         return self._api.request_post('/list-device/value-delta', json={'devices': devices}).check().payload_raw    # type: ignore

#   '''used in services: not used'''
#     @internal_api_error_handler
#     def get_obj_values_deltas_for_group_devices(
#         self,
#         devices_info_box: TDevicesInfoBox,
#         period_from: datetime,
#         period_to: datetime,
#         limit: int,
#         offset: int,
#     ) -> List[Dict[str, Any]]:  # TODO: typing
#         return self._api.request_post('/devices/group/values-deltas', json={"devices_info_box": devices_info_box_to_json(devices_info_box)}, q=get_query_params(
#             period_from=period_from,
#             period_to=period_to,
#             limit=limit,
#             offset=offset,
#         )).check().payload_raw    # type: ignore

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_object_imbalance_by_period(
        self,
        devices: Dict[str, ApiDeviceImbalanceModel],
        period_from: date,
        period_to: date,
    ) -> List[ApiImbalanceValueDeviceListResponse]:
        # Dict["object_id": Dict[
        # "flat_devices":List[List['device_id: UUID', channel: int]],
        # "group_devices":List[List['device_id: UUID', channel: int]]]]
        return self._api_device.request_post(
            '/list-device/imbalance',
            q={'period_from': period_from, 'period_to': period_to},
            json={'devices': devices},
        ).typed(List[ApiImbalanceValueDeviceListResponse]).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_object_actual_imbalance_by_period(
        self,
        devices: Dict[str, ApiDeviceImbalanceModel],
        period_from: date,
        period_to: date,
    ) -> List[ApiImbalanceValueDeviceListResponse]:
        # Dict["object_id": Dict[
        # "flat_devices":List[List['device_id: UUID', channel: int]],
        # "group_devices":List[List['device_id: UUID', channel: int]]]]
        return self._api_device.request_post(
            '/list-device/imbalance/actual',
            q={'period_from': period_from, 'period_to': period_to},
            json={'devices': devices},
        ).typed(List[ApiImbalanceValueDeviceListResponse]).check().payload

#   '''used in services: not used'''
#     @internal_api_error_handler
#     def get_graphics_object_imbalance_by_period(
#         self,
#         devices: Dict[str, List[List[Any]]],
#         period_from: date,
#         period_to: date,
#     ) -> Dict[str, Any]:  # TODO: typing
#         # Dict["name_object": List[List['object_id': UUID]],
#         # "group_devices":List[List['device_id: UUID', channel: int]],
#         # "flat_devices":List[List['device_id: UUID', channel: int]]}
#         return self._api.request_post(
#             '/list-device/imbalance-graphics',
#             q={'period_from': period_from, 'period_to': period_to},
#             json={'devices': devices},
#         ).check().payload_raw    # type: ignore

#     '''used in services: not used'''
#     @internal_api_error_handler
#     def get_events_magnet_devices(
#         self,
#         devices: List[UUID],
#         period_from: date,
#         period_to: date,
#     ) -> Dict[str, Any]:  # TODO: typing
#         return self._api.request_post(
#             '/events-magnet/devices',
#             q={'period_from': period_from, 'period_to': period_to},
#             json=ApiDeviceEvent(devices=devices),
#         ).check().payload_raw    # type: ignore

#     '''used in services: not used'''
#     @internal_api_error_handler
#     def get_events_low_battery_devices(
#         self,
#         devices: List[UUID],
#         period_from: date,
#         period_to: date,
#     ) -> Dict[str, Any]:  # TODO: typing
#         return self._api.request_post(
#             '/events-battery/devices',
#             q={'period_from': period_from, 'period_to': period_to},
#             json=ApiDeviceEvent(devices=devices),
#         ).check().payload_raw    # type: ignore

#     '''used in services: not used'''
#     @internal_api_error_handler
#     def get_events_devices(
#         self,
#         devices: List[UUID],
#         period_from: date,
#         period_to: date,
#     ) -> Dict[str, Any]:  # TODO: typing
#         return self._api.request_post(
#             '/events/devices',
#             q={'period_from': period_from, 'period_to': period_to},
#             json=ApiDeviceEvent(devices=devices),
#         ).check().payload_raw    # type: ignore

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_devices_not_data(
        self,
        devices: ApiDeviceNoDataList,
        limit: Optional[int],
        offset: Optional[int],
        filters: Optional[List[Dict[str, Any]]],
        sorts: Optional[List[Tuple[str, str]]],
    ) -> ApiDeviceNoDataResponse:
        query_params: Dict[str, Any] = {
            "filter": filters,
            "sort": sorts,
            "limit": limit,
            "offset": offset,
        }
        return self._api_device.request_post(
            '/devices/no-data',
            q=query_params,
            json=devices,
        ).typed(ApiDeviceNoDataResponse).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def upload_device(
        self,
        device_id: Optional[UUID],
        manufacturer_name: str,
        mac: int,
        manufacturer_serial_number: str,
        modification_type_id: str,
        modification_id: Optional[str],
        date_produced: Optional[datetime],
        firmware_version: Optional[str],
        hardware_version: Optional[str],
        uplink_protocol_id: UUID,
        downlink_protocol_id: UUID,
        key_id: Optional[UUID],
        uplink_encryption_key: Optional[str],
        downlink_encryption_key: Optional[str],
        uplink_encryption_type: EncryptionType,
        downlink_encryption_type: EncryptionType,
        data_input_gateway_network_id: UUID,
        data_gateway_id: UUID,
        device_channels: List[GeneralApiDeviceChannel],
        modification_name: Optional[str],  # for script
        modification_type_name: Optional[str],  # for script
    ) -> ApiImportDeviceResponse:
        return self._api_device.request_post('/import/devices', json={
            'device_id': str(device_id) if device_id is not None else None,
            'manufacturer_name': manufacturer_name,
            'mac': mac,
            'manufacturer_serial_number': manufacturer_serial_number,
            'modification_type_id': modification_type_id,
            'modification_id': modification_id,
            'date_produced': str(date_produced) if date_produced is not None else None,
            'firmware_version': firmware_version,
            'hardware_version': hardware_version,
            'uplink_protocol_id': str(uplink_protocol_id) if uplink_protocol_id is not None else None,
            'downlink_protocol_id': str(downlink_protocol_id) if downlink_protocol_id is not None else None,
            'key_id': str(key_id) if key_id is not None else None,
            'uplink_encryption_key': uplink_encryption_key,
            'downlink_encryption_key': downlink_encryption_key,
            'uplink_encryption_type': uplink_encryption_type,
            'downlink_encryption_type': downlink_encryption_type,
            'data_input_gateway_network_id': str(data_input_gateway_network_id),
            'data_gateway_id': str(data_gateway_id),
            'device_channels': device_channels,
            'modification_name': modification_name,
            'modification_type_name': modification_type_name,
        }).typed(ApiImportDeviceResponse).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_events_to_sync_api_user_data(self) -> DeviceStateSyncDataResponse:
        return self._api_device.request_get('/device-state-sync/events').typed(DeviceStateSyncDataResponse).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def update_device_state_sync_offset(self, sync_offset_id: UUID, is_confirmed: bool = True) -> UpdateSyncOffsetResponse:
        return self._api_device.request_patch(f'/device-state-sync-offset/{str(sync_offset_id)}', json={'is_confirmed': is_confirmed}).typed(UpdateSyncOffsetResponse).check().payload

#     '''used in services: not used'''
#     @internal_api_error_handler
#     def get_base_station_by_token(self, api_token: Optional[Tuple[Optional[str], str]]) -> Dict[str, Any]:  # TODO: typing
#         assert self._api_base_station is not None
#         return self._api_base_station.request_get('/base-station').check().payload_raw    # type: ignore

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_modifications_for_list_devices(self, devices: GeneralDeviceValueModel) -> ApiModificationResponseForDevices:
        return self._api_device.request_post(
            '/device-modifications/list-devices',
            json=devices,
        ).typed(ApiModificationResponseForDevices).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_last_value_date_by_device(
        self,
        device: DeviceValueDescriptionModel,
        period_selection: Optional[date],
        period_to: Optional[datetime],
    ) -> ApiDeviceLastValueDateResponse:
        return self._api_device.request_post(
            '/devices/last-value-date',
            json=device,
            q={"period_selection": period_selection, "period_to": period_to},
        ).typed(ApiDeviceLastValueDateResponse).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_last_value_date_by_device_channel(
        self,
        device_channel: DeviceChannelValueDescriptionModel,
        period_selection: Optional[date],
        period_to: Optional[datetime],
    ) -> ApiDeviceChannelLastValueDateResponse:
        return self._api_device.request_post(
            '/device-channels/last-value-date',
            json=device_channel,
            q={"period_selection": period_selection, "period_to": period_to},
        ).typed(ApiDeviceChannelLastValueDateResponse).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_last_value_date_for_list_devices(
        self,
        devices: GeneralDeviceValueModel,
        period_selection: Optional[date],
        period_to: Optional[datetime],
    ) -> ApiLastValueDateResponseForDevices:
        return self._api_device.request_post(
            '/device-channels/last-value-date/list-devices',
            json=devices,
            q={"period_selection": period_selection, "period_to": period_to},
        ).typed(ApiLastValueDateResponseForDevices).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_last_value_date_by_device_meter_list(
        self,
        device_meters: GeneralDeviceMeterValueModel,
        period_selection: Optional[date],
        period_to: Optional[datetime],
    ) -> ApiDeviceMeterLastValueDateResponse:
        return self._api_device.request_post(
            '/device-meters/last-value-date/list-meters',
            json=device_meters,
            q={"period_selection": period_selection, "period_to": period_to},
        ).typed(ApiDeviceMeterLastValueDateResponse).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_last_value_date_by_device_meter(
        self,
        device_meter: DeviceMeterValueDescriptionModel,
        period_selection: Optional[date],
        period_to: Optional[datetime],
    ) -> ApiDeviceMeterLastValueDateResponse:
        return self._api_device.request_post(
            '/device-meters/last-value-date',
            json=device_meter,
            q={"period_selection": period_selection, "period_to": period_to},
        ).typed(ApiDeviceMeterLastValueDateResponse).check().payload

    '''used in services: Iot Account Application Backend, Service Data Gateway'''
    @internal_api_error_handler
    def get_data_gateway_networks(
        self,
        data_gateway_id: UUID,
        limit: Optional[int],
        offset: Optional[int],
        filters: Optional[List[Dict[str, Any]]],
        sorts: Optional[List[Tuple[str, str]]],
    ) -> List[ApiDataGatewayNetworkResponse]:
        query_params: Dict[str, Any] = {
            "filter": filters,
            "sort": sorts,
        }
        if limit is not None:
            query_params['limit'] = limit
        if offset is not None:
            query_params['offset'] = offset
        return self._api_device.request_get(
            f'/data-gateways/{data_gateway_id}/networks',
            q=query_params,
        ).typed(List[ApiDataGatewayNetworkResponse]).check().payload

    '''used in services: Iot Account Application Backend'''
    @internal_api_error_handler
    def get_device_modifications(
        self,
        limit: Optional[int],
        offset: Optional[int],
        filters: Optional[List[Dict[str, Any]]],
        sorts: Optional[List[Tuple[str, str]]],
    ) -> List[ApiDeviceModificationResponse]:
        query_params: Dict[str, Any] = {
            "filter": filters,
            "sort": sorts,
        }
        if limit is not None:
            query_params['limit'] = limit
        if offset is not None:
            query_params['offset'] = offset
        return self._api_device.request_get('/device-modifications', q=query_params).typed(List[ApiDeviceModificationResponse]).check().payload

    '''used in services: Iot Account Application Gateway, Service Data Gateway'''
    @internal_api_error_handler
    def get_device_profiles_granularity_list_by_device_id(
        self,
        device_id: UUID,
        date_from: Optional[datetime],
        date_to: Optional[datetime],
        profile_kinds: Optional[str],
    ) -> List[ApiDeviceProfileGranularityResponse]:
        query: Dict[str, Any] = {"date_from": date_from, "date_to": date_to}
        if profile_kinds:
            query.update({"profile_kinds": profile_kinds})
        return self._api_device.request_get(
            f'/devices/{device_id}/device-profiles/granularies',
            q=query,
        ).typed(List[ApiDeviceProfileGranularityResponse]).check().payload

    '''used in services: Iot Account Application Gateway, Service Data Gateway'''
    @internal_api_error_handler
    def get_device_profiles_list(
        self,
        device_id: UUID,
        granularity_s: Optional[str],
        profile_kinds: Optional[str],
        dt_from: datetime,
        dt_to: datetime,
    ) -> List[ApiDeviceProfileResponse]:
        query: Dict[str, Any] = {"dt_from": dt_from, "dt_to": dt_to}
        if profile_kinds:
            query.update({"profile_kinds": profile_kinds})
        if granularity_s:
            query.update({"granularity_s": granularity_s})
        return self._api_device.request_get(
            f'/devices/{device_id}/device-profiles',
            q=query,
        ).typed(List[ApiDeviceProfileResponse]).check().payload

    @internal_api_error_handler
    def get_bs_downlink_tasks(
        self,
    ) -> List[ApiBSDownlinkTaskResponse]:
        assert self._api_device is not None
        return self._api_device.request_get(
            '/bs-downlink-tasks',
        ).typed(List[ApiBSDownlinkTaskResponse]).check().payload

    @internal_api_error_handler
    def get_bs_downlink_task_by_id(
        self,
        task_id: UUID,
    ) -> ApiBSDownlinkTaskResponse:
        assert self._api_device is not None
        return self._api_device.request_get(
            f'/bs-downlink-tasks/{task_id}',
        ).typed(ApiBSDownlinkTaskResponse).check().payload

    @internal_api_error_handler
    def get_binary_data_list(
        self,
    ) -> List[ApiBinaryDataResponse]:
        assert self._api_device is not None
        return self._api_device.request_get(
            '/binary-data',
        ).typed(List[ApiBinaryDataResponse]).check().payload

    @internal_api_error_handler
    def get_binary_data_by_id(
        self,
        binary_data_id: UUID,
    ) -> ApiBinaryDataResponse:
        assert self._api_device is not None
        return self._api_device.request_get(
            f'/binary-data/{binary_data_id}',
        ).typed(ApiBinaryDataResponse).check().payload

#     '''used in services: not used'''
#     @internal_api_error_handler
#     def get_data_gateway_netowork_device_list(
#         self,
#         gateway_id: UUID,
#         devices: List[ApiDataGatewayNetworkDevicesInfo],
#         limit: Optional[int],
#         offset: Optional[int],
#         filters: Optional[List[Dict[str, Any]]],
#         sorts: Optional[List[Tuple[str, str]]],
#     ) -> InternalApiResponse[List[ApiDataGatewayNetworkDeviceInfoResponse]]:        # type: ignore
#         query_params: Dict[str, Any] = {
#             "filter": filters,
#             "sort": sorts,
#         }
#         if limit is not None:
#             query_params['limit'] = limit
#         if offset is not None:
#             query_params['offset'] = offset
#         return self._api_device.request_post(
#             f'/data-gateways/{gateway_id}/devices',
#             q=query_params,
#             json=json.dumps(ApiDataGatewayNetworkDevicesBody(devices=devices).dict()),
#         ).typed(List[ApiDataGatewayNetworkDeviceInfoResponse]).check()

    '''used in services: Iot Account Application Gateway, Service Data Gateway'''
    @internal_api_error_handler
    def get_data_aggregator_api_user(
        self,
        api_user_id: UUID,
    ) -> DataAggregatorApiUserResponse:
        return self._api_device.request_get(
            f'/api-users/{api_user_id}',
        ).typed(DataAggregatorApiUserResponse).check().payload



from datetime import date, datetime
from typing import Dict, Any, Optional, Union


def get_query_params(
    period_from: Optional[Union[datetime, date]],
    period_to: Optional[Union[datetime, date]],
    limit: Optional[int],
    offset: Optional[int],
) -> Dict[str, Any]:

    query_params = dict()
    if period_from is not None:
        query_params.update({"period_from": period_from})

    if period_to is not None:
        query_params.update({"period_to": period_to})

    if limit is not None:
        query_params.update({"limit": limit})  # type: ignore

    if offset is not None:
        query_params.update({"offset": offset})  # type: ignore
    return query_params



from datetime import datetime
from enum import Enum, IntEnum, unique
from ipaddress import IPv4Address
from typing import Optional, List, Tuple, Dict, Any

from ul_api_utils.errors import ValidateApiError
from pydantic import BaseModel, UUID4, root_validator, Field  # pylint: disable=no-name-in-module
from unipipeline.message.uni_message import UniMessage

from data_aggregator_sdk.types.device import NetworkSysTypeEnum
from data_aggregator_sdk.constants.enums import IntegrationV0MessageEvent, IntegrationV0MessageMetaBSChannelProtocol, \
    ProtocolEnum, IntegrationV0MessageErrorType


class ProfileGranulation(Enum):
    """
    Enumeration of time granulations for profile packets
    """
    MINUTE_01 = "MINUTE_01"
    MINUTE_02 = "MINUTE_02"
    MINUTE_03 = "MINUTE_03"
    MINUTE_04 = "MINUTE_04"
    MINUTE_05 = "MINUTE_05"
    MINUTE_06 = "MINUTE_06"
    MINUTE_10 = "MINUTE_10"
    MINUTE_12 = "MINUTE_12"
    MINUTE_15 = "MINUTE_15"
    MINUTE_20 = "MINUTE_20"
    MINUTE_30 = "MINUTE_30"
    MINUTE_60 = "MINUTE_60"

    SECOND_01 = "SECOND_01"
    SECOND_02 = "SECOND_02"
    SECOND_03 = "SECOND_03"
    SECOND_04 = "SECOND_04"
    SECOND_05 = "SECOND_05"
    SECOND_06 = "SECOND_06"
    SECOND_10 = "SECOND_10"
    SECOND_12 = "SECOND_12"
    SECOND_15 = "SECOND_15"
    SECOND_20 = "SECOND_20"
    SECOND_30 = "SECOND_30"


class ProfileKind(Enum):
    """
    Enumeration of profile types:
        energy A-, A+, R-, R+
        full power S
        active power P
        reactive power Q
        voltage U
        current I
        frequency F
        point factor K
    """
    # active generated energy A-
    ENERGY_A_N = 'ENERGY_A_N'
    ENERGY_A_N_A = 'ENERGY_A_N_A'
    ENERGY_A_N_B = 'ENERGY_A_N_B'
    ENERGY_A_N_C = 'ENERGY_A_N_C'
    # active consumed energy A+
    ENERGY_A_P = 'ENERGY_A_P'
    ENERGY_A_P_A = 'ENERGY_A_P_A'
    ENERGY_A_P_B = 'ENERGY_A_P_B'
    ENERGY_A_P_C = 'ENERGY_A_P_C'
    # reactive generated energy R-
    ENERGY_R_N = 'ENERGY_R_N'
    ENERGY_R_N_A = 'ENERGY_R_N_A'
    ENERGY_R_N_B = 'ENERGY_R_N_B'
    ENERGY_R_N_C = 'ENERGY_R_N_C'
    # reactive consumed energy R+
    ENERGY_R_P = 'ENERGY_R_P'
    ENERGY_R_P_A = 'ENERGY_R_P_A'
    ENERGY_R_P_B = 'ENERGY_R_P_B'
    ENERGY_R_P_C = 'ENERGY_R_P_C'
    # full power S
    FULL_POWER_ABC = 'FULL_POWER_ABC'
    FULL_POWER_A = 'FULL_POWER_A'
    FULL_POWER_B = 'FULL_POWER_B'
    FULL_POWER_C = 'FULL_POWER_C'
    FULL_POWER_MIN_ABC = 'FULL_POWER_MIN_ABC'
    FULL_POWER_MIN_A = 'FULL_POWER_MIN_A'
    FULL_POWER_MIN_B = 'FULL_POWER_MIN_B'
    FULL_POWER_MIN_C = 'FULL_POWER_MIN_C'
    FULL_POWER_MAX_ABC = 'FULL_POWER_MAX_ABC'
    FULL_POWER_MAX_A = 'FULL_POWER_MAX_A'
    FULL_POWER_MAX_B = 'FULL_POWER_MAX_B'
    FULL_POWER_MAX_C = 'FULL_POWER_MAX_C'
    # active power P
    ACTIVE_POWER_ABC = 'ACTIVE_POWER_ABC'
    ACTIVE_POWER_A = 'ACTIVE_POWER_A'
    ACTIVE_POWER_B = 'ACTIVE_POWER_B'
    ACTIVE_POWER_C = 'ACTIVE_POWER_C'
    ACTIVE_POWER_MIN_ABC = 'ACTIVE_POWER_MIN_ABC'
    ACTIVE_POWER_MIN_A = 'ACTIVE_POWER_MIN_A'
    ACTIVE_POWER_MIN_B = 'ACTIVE_POWER_MIN_B'
    ACTIVE_POWER_MIN_C = 'ACTIVE_POWER_MIN_C'
    ACTIVE_POWER_MAX_ABC = 'ACTIVE_POWER_MAX_ABC'
    ACTIVE_POWER_MAX_A = 'ACTIVE_POWER_MAX_A'
    ACTIVE_POWER_MAX_B = 'ACTIVE_POWER_MAX_B'
    ACTIVE_POWER_MAX_C = 'ACTIVE_POWER_MAX_C'
    # reactive power Q
    REACTIVE_POWER_ABC = 'REACTIVE_POWER_ABC'
    REACTIVE_POWER_A = 'REACTIVE_POWER_A'
    REACTIVE_POWER_B = 'REACTIVE_POWER_B'
    REACTIVE_POWER_C = 'REACTIVE_POWER_C'
    REACTIVE_POWER_MIN_ABC = 'REACTIVE_POWER_MIN_ABC'
    REACTIVE_POWER_MIN_A = 'REACTIVE_POWER_MIN_A'
    REACTIVE_POWER_MIN_B = 'REACTIVE_POWER_MIN_B'
    REACTIVE_POWER_MIN_C = 'REACTIVE_POWER_MIN_C'
    REACTIVE_POWER_MAX_ABC = 'REACTIVE_POWER_MAX_ABC'
    REACTIVE_POWER_MAX_A = 'REACTIVE_POWER_MAX_A'
    REACTIVE_POWER_MAX_B = 'REACTIVE_POWER_MAX_B'
    REACTIVE_POWER_MAX_C = 'REACTIVE_POWER_MAX_C'
    # voltage U
    VOLTAGE_ABC = 'VOLTAGE_ABC'
    VOLTAGE_A = 'VOLTAGE_A'
    VOLTAGE_B = 'VOLTAGE_B'
    VOLTAGE_C = 'VOLTAGE_C'
    VOLTAGE_MIN_ABC = 'VOLTAGE_MIN_ABC'
    VOLTAGE_MIN_A = 'VOLTAGE_MIN_A'
    VOLTAGE_MIN_B = 'VOLTAGE_MIN_B'
    VOLTAGE_MIN_C = 'VOLTAGE_MIN_C'
    VOLTAGE_MAX_ABC = 'VOLTAGE_MAX_ABC'
    VOLTAGE_MAX_A = 'VOLTAGE_MAX_A'
    VOLTAGE_MAX_B = 'VOLTAGE_MAX_B'
    VOLTAGE_MAX_C = 'VOLTAGE_MAX_C'
    # current I
    CURRENT_ABC = 'CURRENT_ABC'
    CURRENT_A = 'CURRENT_A'
    CURRENT_B = 'CURRENT_B'
    CURRENT_C = 'CURRENT_C'
    CURRENT_MIN_ABC = 'CURRENT_MIN_ABC'
    CURRENT_MIN_A = 'CURRENT_MIN_A'
    CURRENT_MIN_B = 'CURRENT_MIN_B'
    CURRENT_MIN_C = 'CURRENT_MIN_C'
    CURRENT_MAX_ABC = 'CURRENT_MAX_ABC'
    CURRENT_MAX_A = 'CURRENT_MAX_A'
    CURRENT_MAX_B = 'CURRENT_MAX_B'
    CURRENT_MAX_C = 'CURRENT_MAX_C'
    # frequency F
    FREQUENCY_ABC = 'FREQUENCY_ABC'
    FREQUENCY_A = 'FREQUENCY_A'
    FREQUENCY_B = 'FREQUENCY_B'
    FREQUENCY_C = 'FREQUENCY_C'
    FREQUENCY_MIN_ABC = 'FREQUENCY_MIN_ABC'
    FREQUENCY_MIN_A = 'FREQUENCY_MIN_A'
    FREQUENCY_MIN_B = 'FREQUENCY_MIN_B'
    FREQUENCY_MIN_C = 'FREQUENCY_MIN_C'
    FREQUENCY_MAX_ABC = 'FREQUENCY_MAX_ABC'
    FREQUENCY_MAX_A = 'FREQUENCY_MAX_A'
    FREQUENCY_MAX_B = 'FREQUENCY_MAX_B'
    FREQUENCY_MAX_C = 'FREQUENCY_MAX_C'
    # power factor K
    POWER_FACTOR_ABC = 'POWER_FACTOR_ABC'
    POWER_FACTOR_A = 'POWER_FACTOR_A'
    POWER_FACTOR_B = 'POWER_FACTOR_B'
    POWER_FACTOR_C = 'POWER_FACTOR_C'
    POWER_FACTOR_MIN_ABC = 'POWER_FACTOR_MIN_ABC'
    POWER_FACTOR_MIN_A = 'POWER_FACTOR_MIN_A'
    POWER_FACTOR_MIN_B = 'POWER_FACTOR_MIN_B'
    POWER_FACTOR_MIN_C = 'POWER_FACTOR_MIN_C'
    POWER_FACTOR_MAX_ABC = 'POWER_FACTOR_MAX_ABC'
    POWER_FACTOR_MAX_A = 'POWER_FACTOR_MAX_A'
    POWER_FACTOR_MAX_B = 'POWER_FACTOR_MAX_B'
    POWER_FACTOR_MAX_C = 'POWER_FACTOR_MAX_C'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


class ResourceType(Enum):
    """
    Enumeration of resource types for consumption and generation info of integration message data
    """
    COMMON = 'COMMON'
    ENERGY_ACTIVE = 'ENERGY_ACTIVE'
    ENERGY_REACTIVE = 'ENERGY_REACTIVE'


class CounterType(Enum):
    """
    Enumeration of counter types for consumption and generation info of integration message data
    """
    COMMON = 'COMMON'
    ENERGY_PHASE_A = 'ENERGY_PHASE_A'
    ENERGY_PHASE_B = 'ENERGY_PHASE_B'
    ENERGY_PHASE_C = 'ENERGY_PHASE_C'


class IntegrationV0MessageClock(BaseModel):
    clock_id: int = Field(
        -1,
        title="Clock ID",
        description="Device timestamp clock ID",
    )
    value: datetime = Field(
        title="Clock value",
        description="Device timestamp clock value",
    )


class IntegrationV0MessageUptime(BaseModel):
    channel_id: int = Field(
        1,
        title="Channel ID",
        description="Device channel Id",
    )
    uptime_s: float = Field(
        title="Uptime seconds",
        description="Uptime seconds value",
    )


class IntegrationV0MessageRelay(BaseModel):
    relay_id: int = Field(
        -1,
        title="Relay ID",
        description="Device relay ID",
    )
    value: bool = Field(
        title="Relay state",
        description="Device relay state",
    )


MINUTES_IN_DAY_CONST = 24 * 60
SECONDS_IN_DAY_CONST = MINUTES_IN_DAY_CONST * 60


GRANULATION_TO_SECONDS_DELTA_MAP = {
    ProfileGranulation.MINUTE_01: 60,
    ProfileGranulation.MINUTE_02: 120,
    ProfileGranulation.MINUTE_03: 180,
    ProfileGranulation.MINUTE_04: 240,
    ProfileGranulation.MINUTE_05: 300,
    ProfileGranulation.MINUTE_06: 360,
    ProfileGranulation.MINUTE_10: 600,
    ProfileGranulation.MINUTE_12: 720,
    ProfileGranulation.MINUTE_15: 900,
    ProfileGranulation.MINUTE_20: 1200,
    ProfileGranulation.MINUTE_30: 1800,
    ProfileGranulation.MINUTE_60: 3600,

    ProfileGranulation.SECOND_01: 1,
    ProfileGranulation.SECOND_02: 2,
    ProfileGranulation.SECOND_03: 3,
    ProfileGranulation.SECOND_04: 4,
    ProfileGranulation.SECOND_05: 5,
    ProfileGranulation.SECOND_06: 6,
    ProfileGranulation.SECOND_10: 10,
    ProfileGranulation.SECOND_12: 12,
    ProfileGranulation.SECOND_15: 15,
    ProfileGranulation.SECOND_20: 20,
    ProfileGranulation.SECOND_30: 30,
}

assert len(GRANULATION_TO_SECONDS_DELTA_MAP.keys()) == len(ProfileGranulation)

GRANULATION_LENGTH_MAP = {
    ProfileGranulation.MINUTE_01: MINUTES_IN_DAY_CONST,
    ProfileGranulation.MINUTE_02: MINUTES_IN_DAY_CONST / 2,
    ProfileGranulation.MINUTE_03: MINUTES_IN_DAY_CONST / 3,
    ProfileGranulation.MINUTE_04: MINUTES_IN_DAY_CONST / 4,
    ProfileGranulation.MINUTE_05: MINUTES_IN_DAY_CONST / 5,
    ProfileGranulation.MINUTE_06: MINUTES_IN_DAY_CONST / 6,
    ProfileGranulation.MINUTE_10: MINUTES_IN_DAY_CONST / 10,
    ProfileGranulation.MINUTE_12: MINUTES_IN_DAY_CONST / 12,
    ProfileGranulation.MINUTE_15: MINUTES_IN_DAY_CONST / 15,
    ProfileGranulation.MINUTE_20: MINUTES_IN_DAY_CONST / 20,
    ProfileGranulation.MINUTE_30: MINUTES_IN_DAY_CONST / 30,
    ProfileGranulation.MINUTE_60: MINUTES_IN_DAY_CONST / 60,

    ProfileGranulation.SECOND_01: SECONDS_IN_DAY_CONST,
    ProfileGranulation.SECOND_02: SECONDS_IN_DAY_CONST / 2,
    ProfileGranulation.SECOND_03: SECONDS_IN_DAY_CONST / 3,
    ProfileGranulation.SECOND_04: SECONDS_IN_DAY_CONST / 4,
    ProfileGranulation.SECOND_05: SECONDS_IN_DAY_CONST / 5,
    ProfileGranulation.SECOND_06: SECONDS_IN_DAY_CONST / 6,
    ProfileGranulation.SECOND_10: SECONDS_IN_DAY_CONST / 10,
    ProfileGranulation.SECOND_12: SECONDS_IN_DAY_CONST / 12,
    ProfileGranulation.SECOND_15: SECONDS_IN_DAY_CONST / 15,
    ProfileGranulation.SECOND_20: SECONDS_IN_DAY_CONST / 20,
    ProfileGranulation.SECOND_30: SECONDS_IN_DAY_CONST / 30,
}

assert len(GRANULATION_LENGTH_MAP.keys()) == len(ProfileGranulation)


class IntegrationV0MessageProfile(BaseModel):
    type: ProfileKind = Field(
        title="Profile type",
        description="Profile type is like side view of collecting data for device (ex. 'VOLTAGE_ABC', 'ENERGY_A_N')",
    )
    granulation: ProfileGranulation = Field(
        title="Profile granulation",
        description="Granulation is detailing of collecting data, resolution of graph by device (ex. 1 hour (MINUTE_60), 1 second (SECONDS_01))",
    )
    values: Tuple[Optional[float], ...] = Field(
        title="Profile values",
        description="Sequence of collected values for profile packet",
    )

    @root_validator
    def validate_profile(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        granulation = values['granulation']
        len_values = len(values['values'])

        if GRANULATION_LENGTH_MAP[granulation] == len_values:
            return values
        raise ValidateApiError(msg=f"Incorrect profile type and granulation: Granulation is {granulation}, Length values={len_values}")


class IntegrationV0MessageConsumption(BaseModel):
    tariff: int = Field(
        -1,
        title="Consumption tariff",
        description="Consumption tariff number",
    )
    counter_type: CounterType = Field(
        title="Counter type",
        description="Consumption counter type",
    )
    resource_type: ResourceType = Field(
        title="Resource type",
        description="Consumption resource type",
    )
    channel: int = Field(
        title="Channel",
        description="Consumption channel number for collecting info",
    )
    value: float = Field(
        title="Channel value",
        description="Consumption channel value",
    )
    overloading_value: Optional[float] = Field(
        None,
        title="Overloading channel value",
        description="Consumption channel overloading value. "
                    "If overloading_value = None - value is fully precise - must replace current. if not None - overload potentially could be",
    )


class IntegrationV0MessageGeneration(BaseModel):
    tariff: int = Field(
        -1,
        title="Generation tariff",
        description="Generation tariff number",
    )
    counter_type: CounterType = Field(
        title="Counter type",
        description="Generation counter type",
    )
    resource_type: ResourceType = Field(
        title="Resource type",
        description="Generation resource type",
    )

    value: float = Field(
        title="Generation value",
        description="Generation channel value",
    )
    overloading_value: Optional[float] = Field(
        None,
        title="Overloading value",
        description="Generation overloading value. "
                    "If overloading_value = None - value is fully precise - must replace current. if not None - overload potentially could be",
    )


@unique
class BatteryId(IntEnum):
    """
    Battery Ids enumeration for identification of battery module meter
    """
    COMMON = -1
    # RESERVED = 0
    CAPACITOR = 1
    RADIO_MODULE = 2


class IntegrationV0MessageCurrentBatteryLevel(BaseModel):
    battery_id: int = Field(
        BatteryId.COMMON,
        title="Battery ID",
        description="Device battery ID",
    )
    voltage: float = Field(
        title="Battery voltage",
        description="Device battery voltage value",
    )


class IntegrationV0MessageCurrentTemperature(BaseModel):
    value: float = Field(
        title="Temperature value",
        description="Float value of environment current temperature",
    )


class IntegrationV0MessageMetaNbFiBS0(BaseModel):
    station_id: int = Field(
        title="Station ID",
        description="Station unique identifier",
    )
    modem_id: int = Field(
        title="Modem ID",
        description="Modem unique identifier",
    )
    encrypted: bool = Field(
        title="Channel encryption",
        description="Index availability of channel encryption",
    )
    freq_channel: int = Field(
        title="Channel frequency",
        description="Working channel frequency",
    )
    freq_expect: int = Field(
        title="Expected frequency",
        description="Expected channel frequency",
    )
    message_id: int = Field(
        title="Message ID",
        description="Message unique identifier",
    )
    nbfi_f_ask: int = Field(
        title="NBFI frequency ASK",
        description="NBFI frequency ASK",
    )
    nbfi_iterator: int = Field(
        title="NBFI iterator",
        description="NBFI iterator",
    )
    nbfi_multi: int = Field(
        title="NBFI multi",
        description="NBFI multi",
    )
    nbfi_system: int = Field(
        title="NBFI system",
        description="NBFI system",
    )
    signal_rssi: int = Field(
        title="RSSI signal level",
        description="RSSI signal level of antenna",
    )
    signal_snr: int = Field(
        title="SNR signal level",
        description="SNR signal level",
    )
    time_detected: int = Field(
        title="Time detection",
        description="Time when signal was detected",
    )
    time_published: int = Field(
        title="Time publishing value",
        description="Time when signal was published",
    )
    ul_phy: int = Field(
        title="UL phy",
        description="UL phy",
    )
    baudrate: int = Field(
        -1,  # -1 for old protocols
        title="Baud rate",
        description="Baud rate of channel",
    )
    sdr: int = Field(
        -1,  # -1 for old protocols
        title="SDR number",
        description="SDR numeric number",
    )
    message_type: IntegrationV0MessageMetaBSChannelProtocol = Field(
        IntegrationV0MessageMetaBSChannelProtocol.nbfi,
        title="Message type",
        description="Message protocol type",
    )


class IntegrationV0MessageMetaBSHttp(BaseModel):
    freq: int = Field(
        title="Frequency",
        description="Frequency",
    )
    freq_channel: int = Field(
        title="Frequency channel",
        description="Channel frequency number",
    )
    sdr: int = Field(
        title="SDR number",
        description="SDR number",
    )
    baud_rate: int = Field(
        title="Baud rate",
        description="Baud rate of channel",
    )
    rssi: int = Field(
        title="RSSI value",
        description="RSSI signal level of antenna",
    )
    snr: int = Field(
        title="SNR value",
        description="SNR signal level",
    )

    mac: int = Field(
        title="MAC",
        description="MAC number",
    )
    station_id: UUID4 = Field(
        title="Station ID",
        description="Station unique identifier",
    )
    station_serial_number: int = Field(
        title="Station serial number",
        description="Station serial number",
    )

    dt_detected: datetime = Field(
        title="Detection datetime",
        description="Datetime of detection",
    )
    dt_published: datetime = Field(
        title="Publishing datetime",
        description="Datetime of publishing",
    )
    channel_protocol: IntegrationV0MessageMetaBSChannelProtocol = Field(
        title="Channel protocol",
        description="Channel protocol",
    )


class IntegrationV0MessageMetaNBIoT(BaseModel):
    modem_id: Optional[int] = Field(
        None,
        title="Modem ID",
        description="Modem identifier",
    )
    ip_address: str = Field(
        title="IP address",
        description="IP address of modem device",
    )
    port: int = Field(
        title="Port",
        description="Pointer to network connection",
    )


class IntegrationV0MessageData(BaseModel):
    is_valid: bool = Field(
        True,  # TODO: remove default
        title="Valid message",
        description="Validation integration message data field",
    )
    dt: Optional[datetime] = Field(  # TODO: remove optional
        None,  # TODO: remove default
        title="Datetime message",
        description="Datetime integration message field which shows when message was received",
    )
    battery: List[IntegrationV0MessageCurrentBatteryLevel] = Field(
        default_factory=list,
        title="Battery info",
        description="Structure which contains battery ID and voltage level for current device",
    )
    consumption: List[IntegrationV0MessageConsumption] = Field(
        default_factory=list,
        title="Consumption info",
        description="Structure which contains consumption info for current device",
    )
    generation: List[IntegrationV0MessageGeneration] = Field(
        default_factory=list,
        title="Generation info",
        description="Structure which contains generation info for current device",
    )
    profiles: List[IntegrationV0MessageProfile] = Field(
        default_factory=list,
        title="Profiles info",
        description="Structure which contains profiles info for current device",
    )
    events: List[IntegrationV0MessageEvent] = Field(
        default_factory=list,
        title="Events list",
        description="Events list which contains triggered events info for current device",
    )
    temperature: List[IntegrationV0MessageCurrentTemperature] = Field(
        default_factory=list,
        title="Temperature info",
        description="Structure which contains temperature info for current device",
    )
    relay: List[IntegrationV0MessageRelay] = Field(
        default_factory=list,
        title="Relay info",
        description="Structure which contains relay info for current device",
    )
    clock: List[IntegrationV0MessageClock] = Field(
        default_factory=list,
        title="Clock info",
        description="Structure which contains clock info for current device",
    )
    uptime: List[IntegrationV0MessageUptime] = Field(
        default_factory=list,
        title="Uptime info",
        description="Structure which contains uptime info for current device",
    )


class IntegrationMessageMetaIotAccountGateway(BaseModel):
    network_id: UUID4 = Field(
        title="Network ID",
        description="Network identifier",
    )
    gateway_id: UUID4 = Field(
        title="Gateway ID",
        description="Gateway identifier",
    )
    device_id: Optional[UUID4] = Field(
        None,
        title="Device ID",
        description="Device identifier",
    )
    protocol_id: Optional[UUID4] = Field(
        None,
        title="Protocol ID",
        description="Protocol identifier",
    )


class IntegrationMessageMetaRecycle(BaseModel):
    date: datetime = Field(
        title="Datetime recycle",
        description="Datetime field which contains packet recycle date",
    )


class IntegrationMessageMetaUniversalDataInput(BaseModel):
    user_id: UUID4 = Field(
        ...,
        title="Sender user identifier",     # ApiUser.id
        description="Uniq UUID of user",
    )    # DA token id
    name: str = Field(
        ...,
        title="Sender user name",       # ApiUser.name
        description="Name of user",
    )
    type: str = Field(
        ...,
        title="Sender type",        # ApiUser.type
        description="Name of sender type",
    )
    current_dt: datetime = Field(
        ...,
        title="Current sender date and time",
        description="The current date and time of the sender at the time of sending",
    )
    uptime_s: int = Field(
        ...,
        title="Count uninterrupted uptime seconds",
        description="Counter seconds of sender correct working time",
    )
    geo_longitude: float = Field(
        ...,
        title="Geographical longitude of sender",
        description="Geographical latitude coordinate at the time of sending",
    )
    geo_latitude: float = Field(
        ...,
        title="Geographical latitude of sender",
        description="Geographical latitude coordinate at the time of sending",
    )
    version: str = Field(
        ...,
        title="Sender version",
        description="Sender software / hardware version",
    )
    note: str = Field(
        ...,
        title="Note about sender",
        description="Any usefull information about sender",
    )
    ipv4: IPv4Address = Field(
        ...,
        title='IPv4 sender address',
        description="A unique numerical identifier for every sender that send through the internet",
    )


class IntegrationV0MessageMeta(BaseModel):
    nbfi_bs0: Optional[IntegrationV0MessageMetaNbFiBS0] = Field(
        None,
        title="NBFi BS0 meta message structure",
        description="NBFi BS0 meta message structure",
    )
    nbiot: Optional[IntegrationV0MessageMetaNBIoT] = Field(
        None,
        title="NBIoT meta message",
        description="NBIoT meta message structure",
    )
    bs_http: Optional[IntegrationV0MessageMetaBSHttp] = Field(
        None,
        title="BS HTTP meta message",
        description="BS HTTP meta message structure",
    )
    universal_data_input: Optional[IntegrationMessageMetaUniversalDataInput] = Field(
        None,
        title="Universal data input meta",
        description="Universal data input meta structure",
    )
    iot_account_gateway: Optional[IntegrationMessageMetaIotAccountGateway] = Field(
        None,
        title="IoT account gateway",
        description="IoT account gateway structure",
    )
    recycle: Optional[IntegrationMessageMetaRecycle] = Field(
        None,
        title="Recycle info",
        description="Structure which contains integration message recycle date",
    )

    @root_validator()
    def check_meta_exists(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if not any((
            values.get('nbfi_bs0'),
            values.get('nbiot'),
            values.get('bs_http'),
            values.get('universal_data_input'),
            values.get('iot_account_gateway'),
            values.get('recycle'),
        )):
            raise ValueError('one of meta sctructure must be set')
        return values


class IntegrationV0MessageError(BaseModel):
    error_message: str = Field(
        '',
        title="Error message",
        description="String with error description",
    )
    error_type: IntegrationV0MessageErrorType = Field(
        IntegrationV0MessageErrorType.none,
        title="Error type",
        description="This field contains error group name",
    )


class IntegrationV0MessageGateway(UniMessage):
    data: IntegrationV0MessageData = Field(
        title="Main data structure",
        description="Main data structure of message consists of some data structures with device load info such as battery, consumption, temperature, events and etc",
    )
    meta: IntegrationV0MessageMeta = Field(
        title="Meta data structure",
        description="Meta data structure of message consists of some data structures with some additional info for instance of device",
    )
    packet_type_name: str = Field(
        '',
        title="Packet type name",
        description="Each message contains packet type information",
    )
    id: Optional[UUID4] = Field(
        None,
        title="Message ID",
        description="Message ID which is UniMessage.id",
    )
    device_mac: int = Field(
        title="Device MAC",
        description="Each message contains device MAC info",
    )
    raw_message: str = Field(
        title="Raw message string",
        description="Just byte string of message",
    )
    raw_payload: str = Field(
        title="Raw payload string",
        description="Just string of unparsed payload without channel protocol wrapper",
    )
    decrypted_payload: str = Field(
        title="Decrypted payload string",
        description="Just string of unparsed but decrypted payload without channel protocol wrapper."
                    "If no encryption protocol, then this field equals 'raw_payload'",
    )
    device_raw_mac: int = Field(
        title="Device raw MAC",
        description="Each message contains device MAC info raw integer field",
    )

    protocol_id: Optional[UUID4] = Field(
        None,
        title="Protocol ID",
        description="Protocol ID for the payload",
    )
    protocol_name: Optional[str] = Field(
        None,
        title="Protocol name string",
        description="Common packet protocol name info of payload",
    )
    protocol_type: Optional[ProtocolEnum] = Field(
        None,
        title="Protocol type string",
        description="Common packet protocol type info of payload",
    )
    network_id: Optional[UUID4] = Field(
        None,
        title="Network ID",
        description="Network ID info",
    )
    network_sys_type: Optional[NetworkSysTypeEnum] = Field(
        None,
        title="Network system type info",
        description="Network system type info",
    )

    date_created: datetime = Field(
        title="Message date creation",
        description="The date when message was created",
    )
    gateway_id: UUID4 = Field(
        title="Gateway ID",
        description="Gateway ID field info",
    )
    device_id: Optional[UUID4] = Field(
        None,
        title="Device ID",
        description="Device ID field info",
    )
    verified_device: bool = Field(
        title="Device verification info",
        description="This field shows that device was found in DB and contains some data",
    )

    dt_calculated: Optional[bool] = Field(
        False,
        title="Datetime calculation info",
        description="Datetime calculation info",
    )
    error: Optional[IntegrationV0MessageError] = Field(
        None,
        title="Error info",
        description="Error structure contains error type and error message string",
    )



import pytest
from pydantic import ValidationError

from data_aggregator_sdk.integration_message import IntegrationV0MessageProfile, ProfileKind, ProfileGranulation

INTEGRATION_MESSAGE = IntegrationV0MessageProfile(type=ProfileKind.ENERGY_A_P, granulation=ProfileGranulation.MINUTE_60, values=(1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None))  # noqa


def test_integration_message_validation() -> None:
    assert INTEGRATION_MESSAGE != IntegrationV0MessageProfile(type=ProfileKind.ENERGY_A_P, granulation=ProfileGranulation.MINUTE_30, values=(None, None, None, None, None, None, None, None, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, None, None, None, None, None, None, None, None))  # noqa

    with pytest.raises(ValidationError):
        IntegrationV0MessageProfile(type=ProfileKind.ENERGY_A_P, granulation=ProfileGranulation.MINUTE_60, values=(None, None, None, None, None, None, None, None, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1))



from os import path
from setuptools import setup, find_packages

HERE = path.abspath(path.dirname(__file__))

with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='ul-data-aggregator-sdk',
    version='8.10.3',
    description='Data aggregator sdk',
    author='Unic-lab',
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_data={
        '': ['*.yml'],
        'data_aggregator_sdk': ['py.typed'],
    },
    packages=find_packages(include=['data_aggregator_sdk*']),
    include_package_data=True,
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
    ],
    platforms='any',
    install_requires=[
        'requests>=2.26.0',
        'unipipeline>=1.4.3',
        'ul-api-utils>=7.2.8',
        'wtforms==3.0.1',
        'wtforms-alchemy==0.18.0',
        # "ul-api-utils==7.2.8",
        # 'ul-pyncp==1.0.5',
        # 'ul-pysmp==1.0.3',
        # 'ul-data-gateway-sdk==0.4.5',
        # 'ul-api-utils==7.2.7',
        # 'ul-py-tool==1.15.20',
        # 'ul-db-utils==2.10.7'
    ],
)



from uuid import UUID
from datetime import datetime
from typing import Any, Dict, List, Optional

from pydantic import UUID4, BaseModel
from ul_api_utils.api_resource.api_response_payload_alias import ApiBaseUserModelPayloadResponse, JsonApiResponsePayload

from data_aggregator_sdk.constants.enums import DeviceModificationTypeEnum, NetworkSysTypeEnum, NetworkTypeEnum, ProtocolEnum
from data_aggregator_sdk.types.device import ApiDeviceMeterPayloadResponse


# REQUEST BODY


class ApiDataGatewayNetworkDevicesInfo(BaseModel):
    mac: int
    network_id: UUID


class ApiDataGatewayNetworkDevicesBody(BaseModel):
    devices: List[ApiDataGatewayNetworkDevicesInfo]


# RESPONSE BODY


class ApiDeviceManufacturerResponse(JsonApiResponsePayload):
    id: UUID4
    date_created: datetime
    date_modified: datetime
    name: str


class ApiDataGatewayResponse(ApiBaseUserModelPayloadResponse):
    name: str


class ApiProtocolResponse(JsonApiResponsePayload):
    id: UUID4
    date_created: datetime
    date_modified: datetime
    name: str
    type: ProtocolEnum


class ApiDeviceMeteringTypeResponse(JsonApiResponsePayload):
    id: UUID4
    date_created: datetime
    date_modified: datetime

    sys_name: str
    name_ru: str
    name_en: str


class ApiDeviceModificationTypeResponse(JsonApiResponsePayload):
    id: UUID4
    date_created: datetime
    date_modified: datetime

    sys_name: str
    name_ru: str
    name_en: str
    type: DeviceModificationTypeEnum
    metering_type_id: Optional[UUID4]
    device_metering_type: Optional[ApiDeviceMeteringTypeResponse]


class ApiDeviceModificationResponse(JsonApiResponsePayload):
    id: UUID4
    date_created: datetime
    date_modified: datetime
    name: Optional[str]
    device_modification_type_id: Optional[UUID4]
    device_modification_type: ApiDeviceModificationTypeResponse


class ApiDeviceChannelPayloadResponse(ApiBaseUserModelPayloadResponse):
    device_id: UUID4
    serial_number: int
    inactivity_limit: Optional[int]
    device_meter: List[ApiDeviceMeterPayloadResponse]


class ApiDataGatewayNetworkDeviceResponse(ApiBaseUserModelPayloadResponse):
    manufacturer_serial_number: str
    firmware_version: Optional[str]
    hardware_version: Optional[str]
    date_produced: Optional[datetime]
    device_manufacturer: ApiDeviceManufacturerResponse
    device_modification: ApiDeviceModificationResponse
    device_channel: List[ApiDeviceChannelPayloadResponse]


class ApiDataGatewaysNetworkResponse(ApiBaseUserModelPayloadResponse):
    name: str
    type_network: NetworkTypeEnum
    data_gateway_id: UUID4
    data_gateway: ApiDataGatewayResponse
    sys_type: NetworkSysTypeEnum
    specifier: Optional[str]
    params: Optional[Dict[str, Any]]


class ApiDataGatewayNetworkDeviceInfoResponse(JsonApiResponsePayload):
    id: UUID4
    date_created: datetime
    date_modified: datetime
    uplink_protocol_id: UUID4
    downlink_protocol_id: UUID4
    data_gateway_network_id: UUID4
    mac: int
    key_id: Optional[UUID4]
    device_id: UUID4
    device: ApiDataGatewayNetworkDeviceResponse
    uplink_encryption_key: Optional[str]
    uplink_decrypted_encryption_key: Optional[str]
    downlink_encryption_key: Optional[str]
    downlink_decrypted_encryption_key: Optional[str]
    encryption_key: Optional[str]
    protocol: ApiProtocolResponse
    network: Optional[ApiDataGatewaysNetworkResponse]
